program ADFdiff;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  Dialogs,
  ADF_Block_Routines in '..\CORE LIBRARY\ADF_Block_Routines.pas',
  ADF_diff_funcs in '..\CORE LIBRARY\ADF_diff_funcs.pas',
  ADF_file_routines in '..\CORE LIBRARY\ADF_File_Routines.pas',
  CRC32 in '..\CORE LIBRARY\CRC32.pas',
  general_funcs in '..\CORE LIBRARY\general_funcs.pas',
  CompareRoutines in '..\CORE LIBRARY\CompareRoutines.pas',
  SciZipFile in '..\CORE LIBRARY\SciZipFile.pas';

const
  ERRMSG = #13#10 +
           'ADFdiff v1.0 - Written by Big-Byte of the English Amiga Board' +
           #13#10 +
           #13#10 +
           'Usage: adfdiff MAKEDIFF  <master>.adf <slave>.adf.  Output = <slave>.diff' +
           #13#10 +
           '   or: adfdiff MERGEDIFF <master>.adf <slave>.diff. Output = <slave>.adf' +
           #13#10;

var
  exeName     : string;
  command     : string;
  masterFile  : string;
  secondFile  : string;
  outputFile  : string;

  masterCRC   : longword;
  secondCRC   : longword;


procedure MakeDiffFromFile;
var
  masterADF   : TADFlib;
  secondADF   : TADFlib;

  left_b      : byte;
  right_b     : byte;
  i           : integer;
  j           : integer;
  n           : integer;
  cnt         : integer;
  offset      : longint;  // offset value of first byte that is different.
  nbytes      : longint;  // cnt of number of subsequent bytes that are different
  theFile     : file;
  offsetPtr   : Pointer;
  entryList   : TEntryList;

begin

  if not FileExists(masterFile) then
  begin
    writeln('Error: The .adf file: ''' + masterFile + ''' cannot be found');
    exit;
  end;

  if not FileExists(secondFile) then
  begin
    writeln('Error: The .adf file: ''' + secondFile + ''' cannot be found');
    exit;
  end;

  masterADF := TADFlib.Create;
  secondADF  := TADFlib.Create;

  masterADF.loadADF(masterFile, false);
  secondADF.loadADF(secondFile, false);

  masterCRC := masterADF.CRC32;
  secondCRC := secondADF.CRC32;

  //Writeln('CRC of master ADF = ' + IntToHex(masterCRC, 4));
  //Writeln('CRC of second ADF = ' + IntToHex(secondCRC, 4));

  outputFile := copy(secondFile, 1, length(secondFile)-4) + '.diff';


  // Read through both files a byte at a time.
  // Offset = position of first byte of a sequence of bytes that are different
  // nbytes = number of bytes in the current sequence that are different.

  // This code creates a structure in memory as follows:
  //
  // EntryList (TStringList)
  //     TEntry object 1 =  '000001'
  //                            |------- offset value 1
  //                            |------- offset value 2
  //                            |------- etc...
  //     TEntry object 2 =  '000035'
  //     TEntry object 3 =  '000050'
  //     and so on...

  // Each TEntry record has its own TList of offset values. So by looking at the
  // name of the object ie: '000035' and then the list of offsets within that name
  // we can see that at starting at offset nn, the next 450 bytes of data are different.

  cnt     := 0;
  offset  := -1;
  nbytes  := 0;
  entryList := TEntryList.Create;

  while (cnt < 901120) do
  begin
    left_b  := masterADF.getByteFromDiskMap(cnt);
    right_b := secondADF.getByteFromDiskMap(cnt);

    if ((left_b <> right_b) and (offset = -1)) then
      offset := cnt;

    if (left_b <> right_b) then
      nbytes := nbytes + 1;

    if (offset > 0) and (left_b = right_b) then
    begin
      entryList.InsertEntry(nbytes, offset);
      offset := -1;
      nbytes := 0;
    end;
    Inc(cnt);
  end;

  if (offset > 0) then
  begin
    entryList.InsertEntry(nbytes, offset);
    offset := -1;
    nbytes := 0;
  end;


  entryList.sort;

  // Write the sorted data to the diff file.

  // diff file format =
  // First four bytes = number of bytes in this sequence
  // Second four bytes = number of offsets
  // next nnnn bytes = the data
  // repeat...


  AssignFile(theFile, outputFile);
  ReWrite(theFile, 1);
  Seek(theFile, 0);

  // First two longwords of the file are the master/slave adf CRC's
  BlockWrite(theFile, masterCRC, 4);
  BlockWrite(theFile, secondCRC, 4);

  for i := 0 to entryList.count-1 do
  begin
    // Number of bytes in the sequence
    nbytes := StrToInt(entryList.Strings[i]);
    BlockWrite(theFile, nbytes , 4);

    // Number of offsets
    cnt := entryList.GetOffSetListCount(i);
    BlockWrite(theFile, cnt, 4);

    for j := 0 to cnt-1 do
    begin
      n := entryList.getOffsetItemValue(i, j);  // n = offsetvalue

      // Write the offset value...
      BlockWrite(theFile, n, 4);

      // .....followed by the actual bytes
      offsetPtr := secondADF.getDiskMapPtr2(n);
      BlockWrite(theFile, offsetPtr^, nbytes);
    end;
  end;

  CloseFile(theFile);


  masterADF.unLoadADF;
  secondADF.unLoadADF;

  Writeln('Finished! The file: ''' + outputFile + ''' has been written.');

  FreeAndNil(masterADF);
  FreeAndNil(secondADF);

end;


procedure ReMakeFileFromDiff;
var
  masterADF   : TADFlib;

  n       : Longint;
  i     : Integer;

  theFile : file;
  fpos    : longint;
  fsize   : longint;
  nBytes   : integer;
  nOffsets : integer;
  offset   : Longint;

  theBytes : array[0..999] of byte;

  diffFileMasterCRC   : longword;
  diffFileSecondCRC   : longword;

begin

  if not FileExists(masterFile) then
  begin
    writeln('Error: The .adf file: ''' + masterFile + ''' cannot be found');
    exit;
  end;

  if not FileExists(secondFile) then
  begin
    writeln('Error: The .diff file: ''' + secondFile + ''' cannot be found');
    exit;
  end;

  // Load the ADF file into memory.
  masterADF := TADFlib.Create;
  masterADF.loadADF(masterFile, false);

  masterCRC := masterADF.CRC32;

  // Now, read the first longword of the .diff file and see if the master
  // adf is the correct file to apply the diff to.

  AssignFile(theFile, secondFile);    // the .diff file.
  reset(theFile, 1);
  Seek(theFile, 0);

  fpos  := FilePos(theFile);
  fsize := FileSize(theFile);

  BlockRead(theFile, diffFileMasterCRC, 4);   // CRC of Master ADF file.

  // CRC of the file we are about to create. Compare this to the CRC of the file
  // once we have created it to check if it created properly.
  BlockRead(theFile, diffFileSecondCRC, 4);

  if (masterCRC <> diffFileMasterCRC) then
  begin
    Writeln('Wrong ADF file specified. Cannot merge the diff file with the ADF.');
    Writeln('ADF file CRC = ' + IntToHex(masterCRC, 4) + '  Expected: ' +  IntToHex(diffFileMasterCRC, 4));
    exit;
  end;

  // If got to here then everything ok, so read through the diff file and apply the differences to the
  // ADF image we have in memory

  while (fpos < fsize) do
  begin
    BlockRead(theFile, nBytes, 4);
    BlockRead(theFile, nOffsets, 4);

    //Memo2.Lines.add('Number of bytes in the sequence: ' + IntToStr(nBytes));
    //Memo2.Lines.add('Number of Offsets = ' + IntToStr(nOffsets));

    for n := 0 to nOffsets-1 do
    begin
      BlockRead(theFile, offset, 4);
      BlockRead(theFile, theBytes, nBytes);

      // Update the masterADF diskMap with the diff changes.
      for i := 0 to nBytes-1 do
      begin
        // Write Byte value to the memory position
        masterADF.setByteInDiskMap(offset, theBytes[i]);
        Inc(offset);
      end;
      fpos := FilePos(theFile);
    end;
  end;
  CloseFile(theFile);   // Close the .diff file.

  // Generate the new filename from the diff file.
  outputFile := copy(secondFile, 1, length(secondFile)-5) + '.adf';
  masterADF.SaveADF(outputFile, 901120);

  // Finally close the original ADF file
  masterADF.unLoadADF;
  FreeAndNil(masterADF);

  Writeln('Finished! The file: ''' + outputFile + ''' has been written.');



end;



// =============================================================================
// MAIN ENTRY POINT INTO THE FUNCTION
// =============================================================================
begin
  { TODO -oUser -cConsole Main : Insert code here }

  if ParamCount <> 3 then
  begin
    Writeln(ERRMSG);
    exit;
  end;


  exeName     := ParamStr(0);
  command     := ParamStr(1);
  masterFile  := ParamStr(2);
  secondFile  := ParamStr(3);

  if command = 'MAKEDIFF' then
    MakeDiffFromFile;

  if command = 'MERGEDIFF' then
    ReMakeFileFromDiff;

end.
