unit ADF_Block_Routines;

interface
  uses SysUtils, Dialogs;

type

  {=============================================================================
   This TBlock Object is used exclusively by the ADF object.(ADF_file_routines)
   It is created ONLY when an ADF file is loaded so no checking that an ADF
   image has been loaded is necessary in here - instead it should be done in the
   ADF_file_routines unit.
   This object is destroyed when the ADF image is unloaded.
  =============================================================================}
  TBlock = class(TObject)
    private

    public
      constructor create(p :TObject); overload;

      function isBlockFree(blockNum       :Longword) :Boolean;
      procedure markBlockAs(blockNum      :Longword; freeOrUsed :String);
      function getPrevFreeBlock(blockNum  :Longword) :Longword;
      function getNextFreeBlock(blockNum  :Longword) :Longword;
      function numBlocksFree()                       :Integer;
      function numBlocksUsed()                       :Integer;
      function getBlockSecType(blockNum   :Longword) :String;
      function getBlockChecksum(blockNum  :Longword) :String;
      function calcBlockChecksum(blockNum :Longword) :String;
      function calcChecksum(var X; XSize :longword; CheckSumPos:longword; Method:longword):longword;
      function getFHBpermissions(blockNum :Longword) :String;

      procedure add_entry_to_parentdir(parentBlockNum, blockNum :Longword);

  end;

implementation
  uses general_funcs, ADF_file_routines;


// Global variable used to store the TADFlib object reference passed in
// when this object is created.
var
  theADF    :TADFlib;

{===============================================================================
 FUNCTION:    create
 DESCRIPTION: When block object created, Sets up theADF pointer

 ACCEPTS:     Pointer back to parent that created this object.
 RETURNS:     Nothing
===============================================================================}
constructor TBlock.create(p  :TObject);
begin
  theADF := (p as TADFLib);
end;


{===============================================================================
 FUNCTION:    isBlockFree
 DESCRIPTION: Checks if a block has been used or not

 ACCEPTS:     Block Number
 RETURNS:     TRUE if block is Free / FALSE if block is Used (allocated)
===============================================================================}
function TBlock.isBlockFree(blockNum: longword): Boolean;
var
  arrpos          :Longword;
  diff            :Longword;
  bit             :Longword;
  shifted         :Longword;

  bitMapBlockNum  :Longword;
  fRootBlock      :^TRootBlock;
  fBitMapBlock    :^TBitMapBlock;

begin
  // ***************************************************************************
  // The Root block bm_pages[] array contains pointers to one or more
  // bitmap blocks.
  // Each bitmap block contains details about whether all blocks on the disk
  // are free or allocated (starting with block 2 as blocks 0 and 1 are
  // Bootblock blocks and are ignored)
  // There are 1760 blocks in an adf disk so the bitmap pages array will
  // be 1758 bits long (54 longwords + 30 bits of the 55th longword)
  // If a bit is set then the block is free
  // ***************************************************************************

  // Block 2 is stored at bit 0 so account for the shift
  if (blockNum < 2) or (blockNum > 1760) then
  begin
    showMessage('isBlockFree: blockNum '+IntToStr(blockNum)+' out of range');
    Result := true;
  end
  else
  begin
    // As we are only interested in ADF's then we only need to jump to
    // the bitmap block located in bm_pages[0]
    fRootBlock := theADF.getRootBlockPtr();
    bitMapBlockNum := fRootBlock.bm_pages[0];
    Fliplong(bitMapBlockNum);

    // Read the correct bit for the block Number we want to check
    blockNum := blockNum - 2;
    arrpos := blockNum div 32;          // arrpos = correct longword in list
    diff   := blockNum - (arrpos * 32); // diff = correct bit in longword
    shifted := 1 shl diff;              // Shifted = position of set bit for '&' mask

    fbitMapBlock := theADF.getBitmapBlockPtr(bitMapBlockNum);
    bit := fbitMapBlock.map[arrpos];

    Fliplong(bit);
    bit := bit and shifted;

    if bit = 0 then
      Result := False   // block is Used
    else
      Result := True;   // block is Free
  end;
end;


{===============================================================================
 FUNCTION:    markBlockAs
 DESCRIPTION: Sets/Clears a bit in the bitmap table to mark blocks as Free/Used

 ACCEPTS:     Block Number, 'FREE' or 'USED'
 RETURNS:     Nothing
===============================================================================}
procedure TBlock.markBlockAs(blockNum: longword; freeOrUsed :String);
var
  arrpos          :Longword;
  diff            :Longword;
  lwordval        :Longword;
  shifted         :Longword;

  bitMapBlockNum  :Longword;
  fRootBlock      :^TRootBlock;
  fBitMapBlock    :^TBitMapBlock;

begin
  // ***************************************************************************
  // The Root block bm_pages[] array contains pointers to one or more
  // bitmap blocks.
  // Each bitmap block contains details about whether all blocks on the disk
  // are free or allocated (starting with block 2 as blocks 0 and 1 are
  // Bootblock blocks and are ignored)
  // There are 1760 blocks in an adf disk so the bitmap pages array will
  // be 1758 bits long (54 longwords + 30 bits of the 55th longword)
  // If a bit is set then the block is free
  // ***************************************************************************

  // Block 2 is stored at bit 0 so account for the shift
  if (blockNum < 2) or (blockNum > 1760) then
  begin
    showMessage('isBlockFree: blockNum '+IntToStr(blockNum)+' out of range');
    Exit;
  end
  else
  begin
    // As we are only interested in ADF's then we only need to jump to
    // the bitmap block located in bm_pages[0]
    fRootBlock := theADF.getRootBlockPtr();
    bitMapBlockNum := fRootBlock.bm_pages[0];
    Fliplong(bitMapBlockNum);

    // fbitMapBlock points to the actual bitMapBlock which consists of:
    // 1 longword (checksum) followed by 126 longwords
    fbitMapBlock := theADF.getBitmapBlockPtr(bitMapBlockNum);

    // Read the correct bit for the block Number we want to check
    blockNum := blockNum - 2;
    arrpos := blockNum div 32;          // arrpos = correct longword in list
    diff   := blockNum - (arrpos * 32); // diff = correct bit in longword
    shifted := 1 shl diff;              // Shifted = position of set bit for '&' mask

    lwordval := fbitMapBlock.map[arrpos];
    Fliplong(lwordval);

    if (freeOrUsed = 'USED') then
      shifted := not shifted;

    lwordval := lwordval and shifted;
    FlipLong(lwordval);
    fBitMapBlock.map[arrpos] := lwordval;

    // Now recalculate the checksum for this bitmap block.
    fBitMapBlock.checksum := CalcChecksum(fBitMapBlock^,512,0,1);

  end;
end;


{===============================================================================
 FUNCTION:    getPrevFreeBlock
 DESCRIPTION: Gets the Previous free block before the current one.

 ACCEPTS:     Current Block
 RETURNS:     Block Number of next free block or 0 if no free blocks left.
===============================================================================}
function TBlock.getPrevFreeBlock(blockNum: longword): Longword;
var
  n         :Longword;
  prevFree  :Longword;
begin
  prevFree := 0;

  for n := blockNum-1 downto 2 do
  begin
    if isBlockFree(n) then
    begin
      prevFree := n;
      break;
    end;
  end;
  Result := prevFree;
end;


{===============================================================================
 FUNCTION:    getNextFreeBlock
 DESCRIPTION: Gets the next free block after the current one.

 ACCEPTS:    Current Block
 RETURNS:    Block Number of next free block or 0 if no free blocks left.
===============================================================================}
function TBlock.getNextFreeBlock(blockNum: longword): Longword;
var
  n         :Longword;
  nextFree  :Longword;

begin
  nextFree := 0;

  for n:= blockNum+1 to 1760 do
  begin
    if isBlockFree(n) then
    begin
      nextFree := n;
      break;
    end;
  end;
  Result := nextFree;
end;


{===============================================================================
 FUNCTION:    numBlocksFree
 DESCRIPTION: Returns number of Free blocks in an ADF image

 ACCEPTS:     Nothing
 RETURNS:     Number of FREE blocks
===============================================================================}
function TBlock.numBlocksFree(): Integer;
var
  n          :Integer;
  blocksFree :Integer;

begin
  blocksFree := 0;

  for n := 2 to 1760 do
  begin
    if isBlockFree(n) then
      Inc(blocksFree);
  end;
  Result := blocksFree;
end;


{===============================================================================
 FUNCTION:    numBlocksUsed
 DESCRIPTION: Returns number of Used blocks in an ADF image

 ACCEPTS:     Nothing
 RETURNS:     Number of USED blocks
===============================================================================}
function TBlock.numBlocksUsed(): Integer;
var
  n          :Integer;
  blocksUsed :Integer;

begin
  blocksUsed := 0;

  for n := 2 to 1760 do
  begin
    if (not isBlockFree(n)) then
      Inc(blocksUsed);
  end;
  Result := blocksUsed;
end;


{===============================================================================
 FUNCTION:    getBlockSecType
 DESCRIPTION: Returns sec_type of a block.

              The blocks have sec_type in same position in their data structure:
                  RootBlock
                  UserDirectoryBlock
                  FileHeaderBlock
                  FileExtnBlock

 ACCEPTS:     Block Number
 RETURNS:     Type of Block (Root, File Header, Directory etc..)
===============================================================================}
function TBlock.getBlockSecType(blockNum :Longword): String;
var
  secType          :Longword;
  BlockSecTypePtr  :^TRootblock;

begin

  // Boot block does not have a sec_type
  if (blockNum = 0) or (blockNum = 1) then
  begin
    Result := 'Boot';
    Exit;
  end;

  BlockSecTypePtr := theADF.getFileHeaderBlockPtr(blockNum);
  secType := BlockSecTypePtr.sec_type;
  Fliplong(secType);

  case secType of
      $00000001 : Result := 'Root';
      $00000002 : Result := 'Directory';
      $00000003 : Result := 'Soft Link';
      $00000004 : Result := 'Hard Link';
      $FFFFFFFD : Result := 'File';       // -3
      $FFFFFFFC : Result := 'Hard Link';  // -4
  else
      Result := 'Data';
  end;
end;


{===============================================================================
 FUNCTION:    getBlockChecksum
 DESCRIPTION: Get the current stored checksum of a block - which may be wrong
              if the disk has bad data on it.

              Empty block          - Do not have a checksum.
              T_FFS_DataBlock      - Does not have a checksum.

              TBootBlock (block 0) - checksum starts at 4th byte in of block 0
                                     and is checksum of block 0&1 combined.
                                     Block 1 does not have its own checksum.

              TRootBlock           - checksum starts at 20th byte in
              TUserDirectoryBlock  - checksum starts at 20th byte in
              TFileHeaderBlock     - checksum starts at 20th byte in
              TFileExtnBlock       - checksum starts at 20th byte in
              T_OFS_DataBlock      - checksum starts at 20th byte in

              TBitmapBlock         - Checksum starts at 1st byte of the block

 ACCEPTS:     Block Number (BootBlock = 0, OtherBlock = 2+)
 RETURNS:     Checksum
===============================================================================}
function TBlock.getBlockChecksum(blockNum  :longword) :String;
var
  fBootBlock  :^TBootBlock;
  fOtherBlock :^TRootBlock;
  fCheckSum   :Longword;

begin

  if (blockNum = 0) then
  begin
    // Get bootblock checksum
    fBootBlock := theADF.getBootBlockPtr;
    fCheckSum  := fBootBlock.checksum;
    //Fliplong(fChecksum);
    Result := Format('$%0.8x', [fChecksum]);
    exit;
  end;

  if isBlockFree(blockNum) then
  begin
      Result := 'Block is Free';
      exit;
  end;

  // Get any other block's checksum
  fOtherBlock := theADF.getFileHeaderBlockPtr(blockNum);
  fcheckSum   := fOtherBlock.checksum;
  //Fliplong(fChecksum);
  Result := Format('$%0.8x', [fChecksum]);
end;


{===============================================================================
 FUNCTION:    calcBlockChecksum
 DESCRIPTION: Calculate the checksum of a Block

 ACCEPTS:     Block Number (BootBlock = 0, OtherBlock = 2+)
 RETURNS:     Checksum
===============================================================================}
function TBlock.calcBlockChecksum(blockNum  :Longword) :String;
var
  fCheckSum   :longword;
  p           :pointer;

begin
  if (blockNum = 0) then
  begin
    // Calc bootblock checksum

    // Point to the first byte of data of the block.
    p := theADF.getDiskMapPtr(blockNum);
    fChecksum := CalcChecksum(p^,1024,4,0);

    Result := Format('$%0.8x', [fCheckSum]);
  end
  else
  begin
    // Only calc block checksum if it is used.
    if isBlockFree(blockNum) then
      Result := 'Block is Free'
    else
    begin
      // block is used..

      // checksum offset starts 20 bytes in for all other blocks
      p := theADF.getDiskMapPtr(blockNum);
      fChecksum := CalcChecksum(p^,512,20,1);

      Result := Format('$%0.8x', [fCheckSum]);
    end;
  end;
end;


{===============================================================================
 FUNCTION:    calcChecksum
 DESCRIPTION: Calc Checksum of a Block routine - Written by Feltzkrone

 ACCEPTS:     1. first byte of block of memory to checksum
              2. number of bytes to generate a checksum for
              3. offset to where the checksum is stored in the memory block
              4. method 0 = bootblock, method 1 = other blocks

 RETURNS:     Checksum that a given block SHOULD have
===============================================================================}
function TBlock.calcChecksum(
  var X; XSize:longword; CheckSumPos:longword; Method:longword)  :longword;
var
  newcarry      :longword;
  carry         :longword;
  newchecksum   :longword;
  l0            :longword;
  l1            :longword;
  CSbuf         :array[0..3] of byte;
  Data          :array[0..1023] of byte;

begin
  move(x,Data,xSize);
  fillchar(CSBuf, $04, $00);

  if method=1 then
    CSBuf[3]:=0;

  carry:=0;
  //newcarry:=0;

  for l0:=$00 to (XSize-1) div 4+1 do { +1 for possible left-over }
  begin
    if (l0 = (XSize-1) div 4+1) then
      for l1:=$03 downto $00 do
      begin
        if CSBuf[l1]+carry>=$100 then
           newcarry:=$01
        else
          newcarry:=$00;

        CSBuf[l1]:= CSBuf[l1] + carry;

        if (method=0) or (l1>0) then
          carry:=newcarry
        else
          carry:=0;
      end
    else
      if l0 <> CheckSumPos div 4 then
        for l1:=$03 downto $00 do
        begin
          if CSBuf[l1]+Data[l0*$04+l1]+carry>=$100 then
            newcarry:=$01
          else
            newcarry:=$00;
               CSBuf[l1] := CSBuf[l1] + Data[l0*$04+l1] + carry;
               if (method=0) or (l1>0) then
            carry:=newcarry
          else
            carry:=0;
        end;
      end;

      move(CSBuf,newchecksum,4);
      if method=0 then
        CalcCheckSum:= not NewCheckSum
      else
      begin
        FlipLong(newchecksum);
        newchecksum:=-newchecksum;
        FlipLong(newchecksum);
        CalcCheckSum:=NewCheckSum;
  end;
end;


{===============================================================================
 FUNCTION:    getFHBpermissions
 DESCRIPTION: Get the permissions (READ/WRITE/EXECUTE etc..) of a file

 ACCEPTS:     Block Number (must be a file header block)
 RETURNS:     Permissions (string)
===============================================================================}
function TBlock.getFHBpermissions(blockNum: Longword): String;
var
  cnt       :Longword;
  r         :string;
  bit       :Longword;
  pf        :Longword;
  shifted   :Longword;
  fhb       :^TFileHeaderBlock;

begin

  fhb := theADF.getFileHeaderBlockPtr(blockNum);
  pf := fhb.protect_flags;
  Fliplong(pf);

  r := '';

  for cnt := 0 to 31 do
  begin
    shifted := 1 shl cnt;
    bit := pf and shifted;

    if bit = 0 then
      r := r + '-'
    else

      case cnt of
        0:  r := r + 'D';
        1:  r := r + 'E';
        2:  r := r + 'W';
        3:  r := r + 'R';
        4:  r := r + 'A';
        5:  r := r + 'P';
        6:  r := r + 'S';
        7:  r := r + 'H';

        8:  r := r + 'D';
        9:  r := r + 'E';
        10: r := r + 'W';
        11: r := r + 'R';
        12: r := r + 'D';
        13: r := r + 'E';
        14: r := r + 'W';
        15: r := r + 'R';

        // 16-30 are reserved
        // 31 = SUID (MulituserFS only)
    end;
  end;
  Result := r;
end;


procedure TBlock.add_entry_to_parentdir(parentBlockNum, blockNum: Longword);
var
  n         :Integer;
  dirBlock  :^TUserDirectoryBlock;
begin

  //showMessage(IntToStr(parentBlockNum) + ':' + IntToStr(blockNum));
  dirBlock := theADF.getDirectoryBlockPtr(parentBlockNum);

  for n := 0 to 71 do
  begin
    if (dirBlock.hashtable[n] = 0) then
    begin
      dirBlock.hashtable[n] := blockNum;
      FlipLong(dirBlock.hashtable[n]);
      break;
    end;
  end;

  dirBlock.checksum := CalcChecksum(dirBlock^,512,20,1);

end;

end.
