unit ADF_file_routines;

interface
uses
  SysUtils, Classes, Dialogs, ADF_Block_Routines, SciZipFile, CRC32;


const  
    // 512*11*2*80 = 901120 bytes on a disk

    // There 1760 (901120/512) blocks on a disk
    // 2 blocks are taken up by the BootBlock
    // 1 block is taken up by the RootBlock
    // All the other blocks are used for files etc..

    BSIZE             = 512; // Block/Sector Size (1 sector = 1 block)
    SECTORS_PER_TRACK = 11;
    TRACKS_PER_CYL    = 2;
    NUM_CYLS          = 80;
    BLOCKS_ON_DISK    = 1760;

    ADF_BUFFER_SIZE = BSIZE * SECTORS_PER_TRACK *
                      TRACKS_PER_CYL * NUM_CYLS; // Amount of bytes a whole ADF image takes

type
  TBootBlock = RECORD
    disktype        :array[0..2] of char;     // 'DOS' or 'PFS'   (3 bytes)
    disktype_flags  :byte;                    // OFS, FFS etc..   (1 byte)
    checksum        :longword;                // Block checksum   (4 bytes)
    rootblock       :longword;                // Ptr to RootBlock (4 bytes)
    bootblock_code  :array[0..1011] of byte;  // BootBlock code   (1012 bytes)                                              //       Total      (1024 bytes)
  end;

  TRootBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;    // Unused in Rootblock (val = $00000000)
    high_seq        :longword;    // Unused (val = $00000000)
    hashtable_size  :longword;    // for floppy disk this value = $00000048
    first_data      :longword;    // unused
    checksum        :longword;    // Block checksum
    hashtable       :array[0..71] of longword;  // 72 longwords
    bm_flag         :longword;    // -1 means VALID
    bm_pages        :array[0..24] of longword;  // bitmap blocks pointers
    bm_ext          :longword;    // First bitmap extension block (Hard disk only)
    r_days          :longword;    // Last root alteration date, days since 01/01/1978
    r_mins          :longword;    // Minutes past midnight
    r_ticks         :longword;    // ticks (1/50sec) past last minute
    name_len        :byte;        // Volume name length
    diskname        :array[0..29] of char;
    unused1         :char;
    unused2         :array[0..1] of longword;
    v_days          :longword;    // Date when disk was last written to
    v_mins          :longword;
    v_ticks         :longword;
    c_days          :longword;    // Date when disk was last formatted
    c_mins          :longword;
    c_ticks         :longword;
    next_hash       :longword;    // Unused
    parent_dir      :longword;    // Unused
    extension       :longword;    // FFS = first directory cache block, 0 otherwise
    sec_type        :longword;    // Block secondary type = ST_ROOT (value 1)
  end;

  TUserDirectoryBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;
    unused1         :array[0..2] of longword;
    checksum        :longword;
    hashtable       :array[0..71] of longword;  // 72 longwords
    unused2         :longword;
    user_id         :word;        // User ID
    group_id        :word;        // Group ID
    protect_flags   :longword;    // Protection flags (set to 0 by default)
    unused3         :longword;
    comment_len     :char;        // Directory comment length
    comment         :array[0..78] of char;
    unused4         :array[0..11] of char;
    days            :longword;
    mins            :longword;
    ticks           :longword;
    name_len        :char;        // Directory name length
    dirname         :array[0..29] of char;
    unused5         :char;
    unused6         :array[0..1] of longword;
    next_link       :longword;
    unused7         :array[0..4] of longword;
    hash_chain      :longword;
    parent          :longword;
    extension       :longword;
    sec_type        :longword;
  end;

  TFileHeaderBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;    // Self pointer (to this block)
    high_seq        :longword;    // Number of data block ptr stored here
    data_size       :longword;    // Unused
    first_data      :longword;    // First data block ptr
    checksum        :longword;    // Same algorithm as RootBlock
    data_blocks     :array[0..71] of longword;
    unused1         :longword;
    user_id         :word;        // User ID
    group_id        :word;        // Group ID
    protect_flags   :longword;    // Protection flags (set to 0 by default)
    byte_size       :longword;    // File size in bytes
    comm_len        :byte;        // File comment length (max 79)
    comment         :array[0..78] of char;  // Comment   (max 79 chars)
    unused2         :array[0..11] of char;  // Set to 0
    days            :longword;    // Last change date of file
    mins            :longword;
    ticks           :longword;
    name_len        :byte;        // Filename length
    filename        :array[0..29] of char; // Max 30 chars permitted
    unused3         :char;
    unused4         :longword;
    real_entry      :longword;    // FFS : unused (=0)
    next_link       :longword;    // FFS : hardlinks chained list (first=newest)
    unused5         :array[0..4] of longword;    // set to 0
    hash_chain      :longword;    // Next entry ptr with same hash
    parent          :longword;    // Parent directory
    extension       :longword;    // Pointer to 1st file extension block
    sec_type        :longword;    // Secondary type : ST_FILE (= -3)
  end;

  TFileExtnBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;    // Self pointer (to this block)
    high_seq        :longword;    // Number of data block ptr stored here
    unused1         :longword;    // Unused
    unused2         :longword;    // Unused
    checksum        :longword;    // Same algorithm as RootBlock
    data_blocks     :array[0..71] of longword;
    info            :array[0..45] of longword;    // Unused (== 0)
    unused3         :longword;    // Unused
    parent          :longword;    // Parent file header block
    extension       :longword;    // Next file header extension block (0=this is the last)
    sec_type        :longword;    // Secondary type : ST_FILE (= -3)
  end;

  T_OFS_DataBlock = RECORD
    primtype        :longword;    // Value = $00000008 (T_DATA)
    header_key      :longword;    // Pointer to file header block
    seq_num         :longword;    // File data block number (first is #1)
    data_size       :longword;    // data size (will be <= 512-24)
    next_data       :longword;    // Next data block ptr (0 for last)
    checksum        :longword;    // Same algorithm as RootBlock
    data            :array[0..487] of byte; // 488 bytes data per block
  end;

  T_FFS_DataBlock = Record
    data            :array[0..511] of byte;  // 512 bytes of data per block
  end;

  TBitmapBlock = RECORD
    checksum :longword;
    map      :array[0..126] of longword;
  end;

  // Object to hold details about one entry in the directory list.
  TDirectoryEntry = class(TObject)
    name                :String;
    DirFileRoot         :char;        // 'D'irectory, 'F'ile or 'R'oot
    permissions         :String;
    blockNum            :Longword;
    size_bytes          :longword;
    size_blocks         :longword;
    level               :Integer;
    parentdir_name      :String;
    parentdir_blockNum  :longword;
  end;

  Tcallbackfmt = procedure(s: String) of object; // Used for cbf('xxx') debug log

type
  TADFlib = class(TObject)

  private
    // The whole ADF disk image stored as bytes in here.
    diskMap     :array[0..ADF_BUFFER_SIZE-1] of byte;
    fileMap     :array of byte;

    // Various internal status variables
    fFileName           :String;              // Windows filename
    fNameOfADFinZip     :String;

    fADFloaded          :Boolean;

    blockList           :TStringList;

    // When dumping a file to the fileMap array the first variable
    // contains the number of bytes to save and the second contains
    // the number of bytes currently saved.

    num_bytes_to_save   :longword;
    num_bytes_saved     :longword;

    tot_bytes           :Integer;
    cur_byte            :Integer;
    fFloppySize         :Longint;

    procedure generateDirectoryList(level: Integer; blockNum: Longword);
    function getCRC32: Longword;
    function getFloppySize: Longint;

  public

    dirList             :TStringList; // List of all files AND directories in an ADF
    block               :TBlock;
    zipFile             :TZipFile;

    constructor Create;
    destructor Destroy; override;

    property FileLoaded   :Boolean read fADFloaded;
    property NameOfFileInZip  :String read fNameOfADFInZip;
    property CRC32        :Longword read getCRC32;
    property floppySize   :Longint read getFloppySize;

    procedure loadADF(fname: String; genDirList :boolean);     // Load an ADF file
    procedure unLoadADF;                  // unLoad an ADF file

    function getADFWindowsFileName :String; // Returns 'Windows' filename of ADF
    function getDiskLabel       :String;    // Returns Disk Name/Label
    function getFileSystem      :String;    // Returns Filesystem  (DOS,PFS etc..)
    function getDiskType        :String;    // Returns Disk Type   (OFS,FFS etc..)
    function isBootable         :Boolean;   // Resturns if disk is bootable or not.

    procedure createFileHeaderBlock(cbf             :Tcallbackfmt;
                                    current_block   :Longword;
                                    num_datablocks  :Longword;
                                    fName           :String;
                                    fSize           :Longword;
                                    parentDirName   :String);

    procedure createFileExtnBlock(cbf               :Tcallbackfmt;
                                    current_block   :Longword;
                                    num_datablocks  :Longword;
                                    fName           :String;
                                    fSize           :Longword;
                                    parentBlockNum  :Longword);

    procedure createDataBlocks(cbf :Tcallbackfmt; current_Block :Longword; numBlocks :Integer; theData :array of byte);

    procedure debugDispFHBInfo(cbf :Tcallbackfmt; blockNum :Longword);
    procedure debugDispFEBInfo(cbf :Tcallbackfmt; blockNum :Longword);


    procedure extractFile(blockNum  :longword; blockType  :String);
    procedure SaveFile(fname :string; flen :Integer);
    procedure SaveADF(fname  :string; flen  :Integer);

    // Load a file from windows into memory ready for inserting into an ADF
    procedure insertFileIntoADFimage(cbf :Tcallbackfmt; dirToInsertInto :String; fname: String);

    // Functions that retrieve Directory/File data
    function getParentDirDetails(blockNum: longword): pointer;

    // Worker functions
    function getByteAsChr(pos: longword;  source :String): String;
    function getByteAsHex(pos: longword;  source :String): String;

    function getByteFromDiskMap(pos: longword) : byte;
    function getLongWordFromDiskMap(pos: longword) : Longword;

    // following functions return a pointer to particular structures of data
    function getDiskMapPtr(blockNum         :Integer) : pointer;
    function getBootBlockPtr()                        : pointer;
    function getRootBlockPtr()                        : pointer;
    function getFileHeaderBlockPtr(blockNum :Integer) : pointer;
    function getFileExtnBlockPtr(blockNum   :Integer) : pointer;
    function getOFSDataBlockPtr(blockNum    :Integer) : pointer;
    function getFFSDataBlockPtr(blockNum    :Integer) : pointer;
    function getBitmapBlockPtr(blockNum     :Integer) : pointer;
    function getDirectoryBlockPtr(blockNum  :Integer) : pointer;

    function getDiskMapPtr2(pos : Integer) : pointer;

    procedure setByteInDiskMap(offset: Longword; byteval: byte);

  end;

implementation

uses general_funcs;



{===============================================================================
 Constructor
===============================================================================}
constructor TADFlib.Create;
begin
  unLoadADF;    // Set some initial default values
end;

{===============================================================================
 Destructor
===============================================================================}
destructor TADFlib.Destroy;
begin
  inherited Destroy;
  unLoadADF;
end;

{===============================================================================
 Function:      LoadADF
 Description:   Loads a complete ADF image into diskMap array

 ACCEPTS:       Filename
 RETURNS:       Nothing
===============================================================================}
procedure TADFlib.loadADF(fname: String; genDirList :boolean);
var
  fADF_file       :File;


  i               :Longint;
  strFileContents :String;

begin
  // If an existing file has already been opened then user must either
  // Save or close it before opening this one.

  if (fADFloaded = true) then
  begin
    // Unload previous ADF before loading this new one.
    unLoadADF;
    // Create List to hold info about files in the ADF
    dirList := TStringList.Create;
  end;

  // Create a new zipFile object
  zipFile := TZipFile.Create;         
  // Load new ADF details
  fFileName     := fname;


  if (copy(fFileName, length(fFileName)-3, 4) = '.zip') then
  begin
    // Its a zip file we are loading.

    zipFile.LoadFromFile(fFileName);    // Load the zip into it.
    fNameOfADFInZip := zipFile.Files[0].filename;

    // Return the FIRST (actually the only one in our case) file in the zip back
    // as a very long string
    strFileContents := zipFile.Uncompressed[0];
    fFloppySize := length(strFileContents);

    // Fast abort current file if oversized.
    if (fFloppySize <> ADF_BUFFER_SIZE) then
    begin
      unLoadADF;
      exit;
    end;


    // Copy the string to our byte array
    for i := 0 to getFloppySize-1 do
      diskMap[i] := byte(strFileContents[i+1]);

  end
  else
  begin
    // Loading an ordinary file
    fillchar(diskMap, ADF_BUFFER_SIZE, $00);  // Zero the array
    AssignFile(fADF_file, fname);             // Read the file into memory
    FileMode := fmShareDenyNone;              // Allow others to access file
    reset(fADF_File, 1);                      // Open the file
    fFloppySize := FileSize(fADF_file);       // Get filesize...
    Seek(fADF_file, 0);                       // Seek to beginning of file

    // Dont read past end of buffer size
    if (fFloppySize <> ADF_BUFFER_SIZE) then
    begin
      unLoadADF;
      exit;
    end;

    if (fFloppySize > 0) then
      BlockRead(fADF_file, diskMap, fFloppySize); // Read nnn bytes of data

    CloseFile(fADF_file);
  end;

  fADFloaded := true;

  if (genDirList) then
  begin
    if (dirList <> nil) then
      FreeAndNil(dirList);
    dirList := TStringList.create;

    block := TBlock.create(self);
    generateDirectoryList(0, 880);
  end;
end;

{===============================================================================
 Function:      UnLoadADF
 Description:   UnLoads an ADF image

 ACCEPTS:       Nothing
 RETURNS:       Nothing
===============================================================================}
procedure TADFlib.unLoadADF;
var
  i   :Integer;
begin

  fillchar(diskMap, ADF_BUFFER_SIZE, $00);    // Zero the array
  fADFloaded      := false;
  fFileName       := '';
  fNameOfADFinZip := '';

  // Clear the file structure list
  if dirList <> nil then
  begin
    // Destroy all the objects within the file list
    for i := 0 to dirList.Count - 1 do
      dirList.Objects[i].Free;
    FreeAndNil(dirList);
  end;

  if (fileMap <> nil) then
    fileMap := nil;

  if (block <> nil) then
    block := nil;

  if (zipFile <> nil) then
    FreeAndNil(zipFile);
end;


{===============================================================================
 Function:      getADFWindowsFileName
 Description:   Returns Windows Filename of ADF image

 ACCEPTS:       Nothing
 RETURNS:       Filename
===============================================================================}
function TADFlib.getADFWindowsFileName: String;
begin
  if (fADFloaded) then
    Result := ExtractFileName(fFileName)
  else
    Result := 'No File Loaded';
end;


{===============================================================================
 FUNCTION:    getDiskLabel
 DESCRIPTION: Returns the Amigados name of the disk

 ACCEPTS:     Nothing
 RETURNS:     Nothing
===============================================================================}
function TADFlib.getDiskLabel: String;
var
  rb    :^TRootBlock;

begin
  if (fADFloaded) then
  begin
    rb := getRootBlockPtr;
    Result := rb.diskname;
  end
  else
    Result := 'No File Loaded';
end;

{===============================================================================
 Function:      getFileSystem
 Description:   Returns File System of ADF image (DOS, PFS etc..)

 ACCEPTS:       Nothing
 RETURNS:       File System
===============================================================================}
function TADFlib.getFileSystem: String;
var
  bbptr   :^TBootBlock;
begin
  if (fADFloaded) then
  begin
    bbptr  := getBootBlockPtr;
    Result := bbptr.disktype;
  end
  else
    Result := 'No File Loaded';
end;

{===============================================================================
 FUNCTION:    getDiskType
 DESCRIPTION: Returns type of disk

 ACCEPTS:     Nothing
 RETURNS:     Type of disk
===============================================================================}
function TADFlib.getDiskType: String;
var
  bbptr    :^TBootBlock;
  dsktype   :byte;
begin

  if (fADFloaded) then
  begin
    bbptr := getBootBlockPtr;

    dskType := bbptr.disktype_flags;
    case dskType of
      0 : Result := 'OFS';
      1 : Result := 'FFS';
      2 : Result := 'OFS/INTL';
      3 : Result := 'FFS/INTL';
      4 : Result := 'OFS/DIRC&INTL';
      5 : Result := 'FFS/DIRC&INTL';
    end;
  end
  else
    Result := 'No File Loaded';
end;



function TADFlib.getParentDirDetails(blockNum: longword): pointer;
var
  fileBlock          :^TFileHeaderBlock;
  uDirBlock          :^TUserDirectoryBlock;

  secType            :String;
  parentDirBlockNum  :longword;
begin

  secType := block.getBlockSecType(blockNum);

  if (secType = 'File') then
  begin
    fileBlock := @diskMap[blockNum*512];
    parentDirBlockNum := fileBlock.parent;
  end
  else
  begin
    uDirBlock := @diskMap[blockNum*512];
    parentDirBlockNum := uDirBlock.parent;
  end;

  Fliplong(parentDirBlockNum);

  uDirBlock := @diskMap[parentDirBlockNum*512];

  //if (parentDirBlockNum = 880) then
  //  uDirBlock.dirname  := 'ROOT';

  Result := uDirBlock;
end;


{===============================================================================
 FUNCTION:    extractFile
 DESCRIPTION: Copies a chosen file within the adf image to memory ready for
              saving to the local hard disk.

 ACCEPTS:     Block Number (must be a File Header Block for first time in.)
              Block Type  ('FileHeader' or 'FileExtn')
 RETURNS:     Nothing.
===============================================================================}
procedure TADFlib.extractFile(blockNum: longword; blockType   :String);
var

  debug             :String;
  lb_ofs            :boolean;
  i                 :Integer;
  cnt               :Integer;
  data_blocks_cnt   :Integer;

  fileHdrBlock      :^TFileHeaderBlock;
  fileExtBlock      :^TFileExtnBlock;

  OFS_DataBlock     :^T_OFS_DataBlock;
  FFS_DataBlock     :^T_FFS_DataBlock;

  dataBlockPtr      :longword;
  fileExtnBlockPtr  :longword;

begin
  // For the OFS there are two ways of reading the contents of a file.
  // First by traversing the linked list of data blocks that is pointed
  // to in first_data (offset 16) and then following the pointers in each
  // file data block.
  // The other way of accessing the file data is by using the
  // data_blocks[] table and going backwards through the data blocks
  // listed there and then the File extension blocks.

  // As the FFS doesn't contain extra information in the data blocks
  // (no pointer list, no checksum) the only way of accessing the file
  // contents is by going through the data_blocks[] table and the File
  // extension blocks.

  // To make it easy I will use the same way which works for both OFS
  // and FFS


  if (getDiskType = 'OFS') then
    lb_ofs := true
  else
    lb_ofs := false;


  if (blockType = 'FileHeader') then
  begin
    // Go into here the first time this function is called.
    // If the file takes up more blocks than the File Header block
    // has space for, then need to recursively call this 'extractFile'
    // function but pass in a blockType of 'FileExtn' instead

    fileHdrBlock := @diskMap[blockNum*512];
    num_bytes_to_save := fileHdrBlock.byte_size;
    Fliplong(num_bytes_to_save);

    // Create the dynamic array in this section to hold the data
    setLength(fileMap, num_bytes_to_save);
    num_bytes_saved := 0;

    debug := 'File Headr Block: '+IntToStr(blockNum)+#13#10;

    // Step through the data blocks in REVERSE order
    data_blocks_cnt := 71;
    cnt             := 1;

    repeat
      dataBlockPtr := fileHdrBlock.data_blocks[data_blocks_cnt];
      FlipLong(dataBlockPtr);
      if dataBlockPtr <> 0 then
      begin
        debug := debug + IntToStr(dataBlockPtr) + '  ';

        // Read the 'data' from the datablock and put it in the
        // filemap dynamic array
        if lb_ofs then
        begin
          // Each datablock has 488 bytes of actual data
          OFS_DataBlock := @diskMap[dataBlockPtr*512];
          for i := 0 to 487 do
          begin
            fileMap[num_bytes_saved] := OFS_DataBlock.data[i];
            Inc(num_bytes_saved);
            if (num_bytes_saved > num_bytes_to_save-1) then
              break;
          end;
        end
        else
        begin
          // Each datablock has 512 bytes of actual data
          FFS_DataBlock := @diskMap[dataBlockPtr*512];
          for i := 0 to 511 do
          begin
            fileMap[num_bytes_saved] := FFS_DataBlock.data[i];
            Inc(num_bytes_saved);
            // num_bytes_to_save-1 cause our num_bytes_saved counter
            // starts at 0 and not 1
            if (num_bytes_saved > num_bytes_to_save-1) then
              break;
          end;
        end;
      end;

      if (cnt mod 10) = 0 then
        debug := debug + #13#10;
      Dec(data_blocks_cnt);
      Inc(cnt);
    until (dataBlockPtr = 0) or (data_blocks_cnt < 0);

    // Display Current Block and all its data block pointers
    //showMessage(debug);

    // Now see if the File Header's file extension block has anything in it.
    fileExtnBlockPtr := fileHdrBlock.extension;
    FlipLong(fileExtnBlockPtr);
    if (fileExtnBlockPtr <> 0) then
    begin
      extractFile(fileExtnBlockPtr, 'FileExtn');
    end;
    // Will get to this point right at end once all file extension blocks
    // have been done.
  end;

  // File extension block code is idential to File header block code
  // apart from the fact that we can get at the actual data in here..

  if (blockType = 'FileExtn') then
  begin
    fileExtBlock := @diskMap[blockNum*512];
    debug := 'File Extn Block: '+IntToStr(blockNum)+#13#10;

    // Step through the data blocks in REVERSE order
    data_blocks_cnt := 71;
    cnt             := 1;

    repeat
      dataBlockPtr := fileExtBlock.data_blocks[data_blocks_cnt];
      FlipLong(dataBlockPtr);

      if dataBlockPtr <> 0 then
      begin
        debug := debug + IntToStr(dataBlockPtr) + '  ';

        // Read the 'data' from the datablock and put it in the
        // filemap dynamic array
        if lb_ofs then
        begin
          // Each datablock has 488 bytes of actual data
          OFS_DataBlock := @diskMap[dataBlockPtr*512];
          for i := 0 to 487 do
          begin
            fileMap[num_bytes_saved] := OFS_DataBlock.data[i];
            Inc(num_bytes_saved);
            if (num_bytes_saved > num_bytes_to_save-1) then
              break;
          end;
        end
        else
        begin
          // Each datablock has 512 bytes of actual data
          FFS_DataBlock := @diskMap[dataBlockPtr*512];
          for i := 0 to 511 do
          begin
            fileMap[num_bytes_saved] := FFS_DataBlock.data[i];
            Inc(num_bytes_saved);
            // num_bytes_to_save-1 cause our num_bytes_saved counter
            // starts at 0 and not 1
            if (num_bytes_saved > num_bytes_to_save-1) then
              break;
          end;
        end;
      end;

      if (cnt mod 10) = 0 then
        debug := debug + #13#10;
      Dec(data_blocks_cnt);
      Inc(cnt);
    until (dataBlockPtr = 0) or (data_blocks_cnt < 0);

    // Display Current Block and all its data block pointers
    //showMessage(debug);

    // Now see if the File Extension block's, file extension block has anything in it.
    fileExtnBlockPtr := fileExtBlock.extension;
    FlipLong(fileExtnBlockPtr);
    if (fileExtnBlockPtr <> 0) then
      extractFile(fileExtnBlockPtr, 'FileExtn');
  end;
end;



{===============================================================================
 FUNCTION:    generateDirectoryList
 DESCRIPTION: Gets the list of files within the ADF and puts them in
              the 'dirList'  TStringList
 ACCEPTS:     First call needs directory block number (880)
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.generateDirectoryList(level :Integer; blockNum: Longword);
var
  i                 :Integer;
  nextLevel         :Integer;
  secType           :String;
  nextChainPtr      :longword;
  nextDirPtr        :longword;
  fileBlock         :^TFileHeaderBlock;
  puDirBlock        :^TUserDirectoryBlock;
  uDirBlock         :^TUserDirectoryBlock;
  theDetails        :TDirectoryEntry;


begin

  nextLevel := level + 1;
  secType   := block.getBlockSecType(blockNum);

  {=============================================================================
    Will go in this section first. If it doesnt then something is wrong!!
    Root Block shares similar structure as UserDirectoryBlock
  =============================================================================}
  if (secType = 'Root') then
  begin
    uDirBlock  := @diskMap[blockNum*512];

    theDetails := TDirectoryEntry.Create;
      theDetails.name               := 'ROOT';
      theDetails.DirFileRoot        := 'R';
      theDetails.permissions        := '    ';
      theDetails.blocknum           := blockNum;
      theDetails.size_bytes         := 0;
      theDetails.size_blocks        := 0;
      theDetails.level              := level;
      theDetails.parentdir_name     := '';
      theDetails.parentdir_blockNum := 0;

    // Dont add root details.
    //dirList.AddObject(theDetails.parentdir_name + theDetails.name, theDetails);

    // As we have identified a directory we need to recursively delve into
    // it and retrieve its files and sub-dirs.
    for i := 0 to 71 do
    begin
      nextDirPtr := uDirBlock.hashtable[i];
      FlipLong(nextDirPtr);

      if nextDirPtr <> 0 then
      begin
        //showMessage(Format('$%0.8x', [fBlockNum]));
        generateDirectoryList(nextLevel, nextDirPtr);
      end;
    end;
  end;

  {=============================================================================
    Will go in this section if a proper directory.
  =============================================================================}
  if (secType = 'Directory') then
  begin
    uDirBlock  := @diskMap[blockNum*512];

    theDetails := TDirectoryEntry.Create;
      theDetails.name               := uDirBlock.dirname;
      theDetails.DirFileRoot        := 'D';
      theDetails.permissions        := '    ';
      theDetails.blocknum           := blockNum;
      theDetails.size_bytes         := 0;
      theDetails.size_blocks        := 0;
      theDetails.level              := level;

    puDirBlock := getParentDirDetails(blockNum);

    // Header key unused in Root Block
    if (puDirblock.header_key = 0) then
    begin
      // Parent dir is Root folder
      theDetails.parentdir_name     := 'ROOT';
      theDetails.parentdir_blockNum := puDirBlock.header_key;
    end
    else
    begin
      theDetails.parentdir_name     := puDirBlock.dirname;
      theDetails.parentdir_blockNum := puDirBlock.header_key;
    end;

    dirList.AddObject(theDetails.parentdir_name + theDetails.name, theDetails);

    // As we have identified a directory we need to recursively delve into
    // it and retrieve its files and sub-dirs.
    for i := 0 to 71 do
    begin
      nextDirPtr := uDirBlock.hashtable[i];
      FlipLong(nextDirPtr);

      if nextDirPtr <> 0 then
      begin
        //showMessage(Format('$%0.8x', [fBlockNum]));
        generateDirectoryList(nextLevel, nextDirPtr);
      end;
    end;

    // Is there any other directory with the same hash value as this one?
    nextChainPtr := uDirBlock.hash_chain;
    Fliplong(nextChainPtr);

    if (nextChainPtr <> 0) then
    begin
      generateDirectoryList(level, nextChainPtr);
    end;
  end;

  {=============================================================================
    Will go in this section if a file
  =============================================================================}
  if (secType = 'File') then
  begin
    fileBlock        := @diskMap[blockNum*512];

    theDetails := TDirectoryEntry.Create;
      theDetails.name               := fileBlock.filename;
      theDetails.DirFileRoot        := 'F';
      theDetails.permissions        := '    ';
      theDetails.blocknum           := blockNum;
      theDetails.size_bytes         := fileBlock.byte_size;
      Fliplong(theDetails.size_bytes);
      theDetails.size_blocks        := 0;
      theDetails.level              := level;

    puDirBlock := getParentDirDetails(blockNum);

    // Header key unused in Root Block
    if (puDirBlock.header_key = 0) then
    begin
      // Parent dir is ROOT folder
      theDetails.parentdir_name     := 'ROOT';
      theDetails.parentdir_blockNum := puDirBlock.header_key;
    end
    else
    begin
      theDetails.parentdir_name     := puDirBlock.dirname;
      theDetails.parentdir_blockNum := puDirBlock.header_key;
    end;

    dirList.AddObject(theDetails.parentdir_name + theDetails.name, theDetails);
  end;
end;



{===============================================================================
 FUNCTION:    saveFile
 DESCRIPTION: Save flen bytes in the fileMap memory block to disk

 ACCEPTS:     Filename to save file to, Number of bytes to save
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.SaveFile(fname  :string; flen  :Integer);
var
  myFile    :File;
begin
  AssignFile(myFile, fname);
  ReWrite(myFile, 1);
  BlockWrite(myFile, fileMap[0], flen);
  CloseFile(myFile);
end;


{===============================================================================
 FUNCTION:    saveADF
 DESCRIPTION: Save all data in the dispMap memory block to disk

 ACCEPTS:     Filename to save file to, Number of bytes to save
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.SaveADF(fname  :string; flen  :Integer);
var
  myFile    :File;
begin
  AssignFile(myFile, fname);
  ReWrite(myFile, 1);
  //BlockWrite(myFile, diskMap[0], flen);
  BlockWrite(myFile, diskMap, flen);
  CloseFile(myFile);
end;





{===============================================================================
 FUNCTION:    insertFileIntoADFimage
 DESCRIPTION:

 ACCEPTS:     Address of function outside this unit which can then be called
              within this procedure. (For debug purpose)
              filename to insert.
              name of parent directory to insert under.

 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.insertFileIntoADFimage(
    cbf               :Tcallbackfmt;
    dirToInsertInto   :String;
    fname             :String);

var
  fhb_created     :Boolean;

  p               :Integer;
  n               :Integer;
  vLong           :Longword;

  div_val         :Longword;
  diff            :Integer;
  blocksToInsert  :Integer;

  feb_needed      :Integer; // File Extn Blocks needed. Will be 0 if <= 72 data blocks.
                            // One block per batch 72 data blocks thereafter
  db_needed       :Integer; // Data Blocks Needed.
  tot_needed      :Integer; // Total Blocks needed.

  dt              :String;
  fFile           :File;
  fSize           :Longword;
  fData           :array of byte;


  fhb             :^TFileHeaderBlock;
  feb             :^TFileExtnBlock;
  genericblock    :^TFileHeaderBlock;

  current_block       :Longword;
  parent_block        :Longword;
  prev_created_block  :Longword;

  csum  :pointer;

begin

  // Load the file that we are inserting into memory.
  cbf('Reading file ' +  fname + ' into memory');

  AssignFile(fFile, fname);           // Read the file into memory
  FileMode := fmShareDenyNone;        // Allow others to access file
  reset(fFile, 1);                    // Open the file
  fSize := FileSize(fFile);           // Get filesize...
  SetLength(fData, fSize);
  fillchar(fData[0], fSize, $00);     // Zero the array
  Seek(fFile, 0);                     // Seek to beginning of file
  BlockRead(fFile, fData[0], fSize);  // Read nnn bytes of data
  CloseFile(fFile);

  // Number of bytes
  tot_bytes := fSize;
  cur_byte  := 0;

  // At this point we have the file we want to insert stored in fData
  // Strip Path data from filename
  fname := extractFileName(fname);

  // Work out how many blocks the data is going to take up.
  // We need to add one extra block to this total for the file header block.
  div_val := 1;

  dt := getDiskType;

  if (dt = 'FFS') then
  begin
    cbf('ADF is FFS. Will insert 512 bytes of data per block');
    div_val := 512;
  end;

  if (dt = 'OFS') then
  begin
    div_val := 488;
    cbf('ADF is OFS. Will insert 488 bytes of data per block');
  end;

  // Work out how many data blocks we need.
  db_needed := fSize div div_val;
  diff      := fSize mod div_val;

  if (diff > 0) then
    Inc(db_needed);

  // Work out how many file extn blocks we need.
  // Remeber we dont need a file extension block for the first 72 data blocks
  n := db_needed - 72;
  feb_needed := 0;

  while (n > 0) do
  begin
    Inc(feb_needed);
    n := n - 72;
  end;

  // Work out total number of blocks needed.
  // 1 FHB + nn Data Blocks + nn File Extn Blocks
  tot_needed := 1 + db_needed + feb_needed;

  cbf('File Header Blocks needed = 1');
  cbf('File Extn Blocks needed   = ' + IntToStr(feb_needed));
  cbf('Data Blocks needed        = ' + IntToStr(db_needed));
  cbf('Total Blocks needed       = ' + IntToStr(tot_needed));
  cbf('Blocks free in ADF image  = ' + IntToStr(block.numBlocksFree));

  if (block.numBlocksFree < (tot_needed)) then
  begin
    cbf('Not enough free blocks on the ADF to inset file');
    Exit;
  end;

  // ***************************************************************************
  // Create the file block by block
  // ***************************************************************************
  fhb_created   := false;
  current_block := 2;    // current block is either the FHB or FEB we are creating
  parent_block  := 0;    // parent block is the parent of the current_block


  while (db_needed > 0) do
  begin
    prev_created_block := current_block;  // Will be 2 on first entry.
    current_block := block.getNextFreeBlock(current_block);

    if (db_needed > 72) then
      blocksToInsert := 72
    else
      blocksToInsert := db_needed;

    if (fhb_created = false) then
    begin
      // Create file header block
      createFileHeaderBlock(cbf, current_block, blocksToInsert, fName, fSize, dirToInsertInto);
    end
    else
    begin
      // Create File Extension Block. parent_block will be valid at this point
      // as it will either be the File Header Block or the previously created
      // File Extension Block
      createFileExtnBlock(cbf, current_block, blocksToInsert, fName, fSize, prev_created_block);
    end;

    // Create data blocks to insert into either the FHB or FEB created above
    // Use my genericblock structure as both FHB and FEB have data_blocks array
    // in the same place.
    genericblock := getFileHeaderBlockPtr(current_block);

    cbf('Data blocks left to insert = ' + IntToStr(db_needed));

    blockList := TStringList.Create;

    // Next function populates the blockList.
    createDataBlocks(cbf, current_block, blocksToInsert, fData);

    p := 71; // first datablock pointer
    for n := 0 to blockList.Count-1  do
    begin
      // Add the list of blocks to the file header block
      vLong := StrToInt(blockList.Strings[n]);
      FlipLong(vLong);
      genericblock.data_blocks[p] := vLong;
      Dec(p);
    end;
    FreeAndNil(blockList);

    // At this point we have to decide whether the extension field needs populating
    if (feb_needed > 0) then
    begin
      Dec(feb_needed);
      genericblock.extension := block.getNextFreeBlock(current_block);
      FlipLong(genericblock.extension);
    end
    else
      genericblock.extension := $00000000;

    // Calculate the checksum for this FHB / FEB
    // then Mark the block as Used.

    csum := getDiskMapPtr(current_block);
    genericblock.checksum := block.CalcChecksum(csum^,512,20,1);
    block.markBlockAs(current_block, 'USED');


    if (fhb_created = false) then
    begin
      // At this point we have created a File Header Block by now, so insert the
      // fhb entry into its parent directory.
      fhb := getFileHeaderBlockPtr(current_block);
      parent_block := fhb.parent;
      FlipLong(parent_block);
      block.add_entry_to_parentdir(parent_block, current_block);

      debugDispFHBInfo(cbf, current_block);   // DEBUG - DISPLAY FILE HEADER BLOCK.
    end
    else
    begin
      debugDispFEBInfo(cbf, current_block);   // DEBUG - DISPLAY FILE HEADER BLOCK.
    end;

    // Finally...
    if (fhb_created = false) then
      fhb_created := true;

    db_needed := db_needed - 72;

  end;

  cbf('Finished!');

  // Refresh the directory list of the disk.
  dirList.Clear;
  generateDirectoryList(0, 880);

end;


{===============================================================================
 FUNCTION:    createFileHeaderBlock
 DESCRIPTION:

 ACCEPTS:     Debug function
              block number of parent block (File Header Block or File Extn Block)
              number of blocks to create
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.createFileHeaderBlock(
  cbf             :Tcallbackfmt;
  current_block   :Longword;
  num_datablocks  :Longword;
  fName           :String;
  fSize           :Longword;
  parentDirName   :String);

var
  n               :Integer;
  fComment        :String;
  dl              :TDirectoryEntry;
  fhb             :^TFileHeaderBlock;

begin

  // Create the file header block
  cbf('Creating File Header Block, stored at: ' +IntToStr(current_block));
  fhb := getFileHeaderBlockPtr(current_block);

  fhb.primtype        := $00000002;       // Value = $00000002 (T_HEADER)
  fhb.header_key      := current_block;   // Self pointer (to this block)
  fhb.high_seq        := num_datablocks;
  fhb.data_size       := 0;      // Unused
  fhb.first_data      := block.getNextFreeBlock(current_block); // First data block ptr

  FlipLong(fhb.primtype);
  FlipLong(fhb.header_key);
  FlipLong(fhb.high_seq);
  FlipLong(fhb.first_data);

  //fhb.checksum      := calc'd later (Same algorithm as RootBlock)
  //fhb.data_blocks   := array[0..71] of longword - populated later

  fhb.unused1         := 0;
  fhb.user_id         := 0;      // User ID (????)
  fhb.group_id        := 0;      // Group ID (????)
  fhb.protect_flags   := 0;      // Protection flags (set to 0 by default)
  fhb.byte_size       := fSize;  // File size in bytes
  FlipLong(fhb.byte_size);

  fComment            := 'Inserted by Big-Byte ADF Toolkit';
  fhb.comm_len        := byte(length(fComment)); // File comment length (max 79)

  // COMMENT
  for n := 0 to fhb.comm_len do
    fhb.comment[n] := fComment[n+1];  // n+1 because fComment.length is stored in first byte

  // Filler
  for n := fhb.comm_len to 79 do
    fhb.comment[n] := chr(0);

  fhb.unused2         := '';         // Array of 12 chars
  fhb.days            := 0;          // Last change date of file
  fhb.mins            := 0;
  fhb.ticks           := 0;

  Fliplong(fhb.days);
  Fliplong(fhb.mins);
  Fliplong(fhb.ticks);

  fhb.name_len := byte(length(fname)); // Filename length

  // FILENAME
  for n := 0 to fhb.name_len do
    fhb.filename[n] := fname[n+1];  // has to be n+1 because 'n' contains string length

  // Filler
  for n := fhb.name_len to 29 do
    fhb.filename[n] := chr(0);

  fhb.unused3         := chr(0);
  fhb.unused4         := 0;
  fhb.real_entry      := 0;   // FFS : unused (=0)
  fhb.next_link       := 0;   // FFS : hardlinks chained list (first=newest)
  fhb.unused5[0]      := 0;   // set to 0
  fhb.unused5[1]      := 0;
  fhb.unused5[2]      := 0;
  fhb.unused5[3]      := 0;
  fhb.unused5[4]      := 0;
  fhb.hash_chain      := 0;   // Next entry ptr with same hash

  Fliplong(fhb.real_entry);
  Fliplong(fhb.next_link);
  Fliplong(fhb.hash_chain);

  if (parentDirName = 'ROOT') then
    fhb.parent := 880
  else
  begin
  for n := 0 to dirList.Count-1 do
    begin
      dl := dirList.Objects[n] as TDirectoryEntry;
      if ((dl.DirFileRoot = 'R') or (dl.DirFileRoot = 'D')
      and (dl.name = parentDirName)) then
      begin
        fhb.parent := dl.blockNum;
        break;
      end;
    end;
  end;

  FlipLong(fhb.parent);

  // Pointer to 1st file extension block - may be updated later
  fhb.extension       := 0;
  fhb.sec_type        := $FFFFFFFD;    // Secondary type : ST_FILE (= -3)

  Fliplong(fhb.extension);
  Fliplong(fhb.sec_type);

end;

{===============================================================================
 FUNCTION:    debugDispFHBInfo
 DESCRIPTION:

 ACCEPTS:     Debug function
              block number of parent block (File Header Block or File Extn Block)
              number of blocks to create
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.debugDispFHBInfo(cbf :Tcallbackfmt; blockNum :Longword);
var
  n     :Integer;
  fhb   :^TFileHeaderBlock;

begin

  fhb := getFileHeaderBlockPtr(blockNum);

  cbf('');
  cbf('File Header Block:');
  cbf('Primtype      = ' + formatLong('H', fhb.primtype));
  cbf('Header Key    = ' + formatLong('H', fhb.header_key) +' ('+formatLong('I', fhb.header_key) +')');
  cbf('High Seq      = ' + formatLong('H', fhb.high_seq)   +' ('+formatLong('I', fhb.high_seq)   +')');
  cbf('Data Size     = ' + formatLong('H', fhb.data_size)  +' ('+formatLong('I', fhb.data_size)  +')');
  cbf('First data    = ' + formatLong('H', fhb.first_data) +' ('+formatLong('I', fhb.first_data) +')');
  cbf('Checksum      = ' + formatLong('H', fhb.checksum));

  for n := 0 to 71 do
  begin
    cbf('Data block: '+IntToStr(n)+' = '+formatLong('H', fhb.data_blocks[n])+'('+formatLong('I', fhb.data_blocks[n]) +')');
  end;

  cbf('User ID       = ' + formatWord('H', fhb.user_id));
  cbf('Group ID      = ' + formatWord('H', fhb.group_id));
  cbf('Protect Flags = ' + formatLong('H', fhb.protect_flags));
  cbf('Byte Size     = ' + formatLong('H', fhb.byte_size)  +' ('+formatLong('I', fhb.byte_size)  +')');
  cbf('Comm Length   = ' + formatByte('H', fhb.comm_len)   +' ('+formatByte('I', fhb.comm_len)   +')');
  cbf('Comment       = ' + fhb.comment);
  cbf('Days          = ' + formatLong('I', fhb.days) + getLastChangeDate(fhb.days));
  cbf('Mins          = ' + formatLong('H', fhb.mins)+' ('+formatLong('I', fhb.mins)+')');
  cbf('Ticks         = ' + formatLong('H', fhb.ticks)+' ('+formatLong('I', fhb.ticks)+')');
  cbf('Filename len  = ' + formatByte('H', fhb.name_len)+' ('+formatByte('I', fhb.name_len)+')');
  cbf('Filename      = ' + fhb.filename);
  cbf('real entry    = ' + formatLong('H', fhb.real_entry));
  cbf('next_link     = ' + formatLong('H', fhb.next_link));
  cbf('hash chain    = ' + formatLong('H', fhb.hash_chain));
  cbf('parent        = ' + formatLong('H', fhb.parent) + ' ('+formatLong('I', fhb.parent)+')');
  cbf('extension     = ' + formatLong('H', fhb.extension) + ' ('+formatLong('I', fhb.extension)+')');
  cbf('sec type      = ' + formatLong('H', fhb.sec_type));
  cbf('');

end;


{===============================================================================
 FUNCTION:    createFileExtnBlock
 DESCRIPTION:

 ACCEPTS:     Debug function
              block number of parent block (File Header Block or File Extn Block)
              number of blocks to create
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.createFileExtnBlock(
  cbf             :Tcallbackfmt;
  current_block   :Longword;
  num_datablocks  :Longword;
  fName           :String;
  fSize           :Longword;
  parentBlockNum  :Longword);

var
  n         :Integer;
  feb       :^TFileExtnBlock;
begin

  // Create the file extension block
  cbf('Creating File Extension Block, stored at: ' +IntToStr(current_block));
  feb := getFileExtnBlockPtr(current_block);

  feb.primtype        := $00000002;       // Value = $00000002 (T_HEADER)
  feb.header_key      := current_block;   // Self pointer (to this block)
  feb.high_seq        := num_datablocks;
  feb.unused1         := 0;    // Unused
  feb.unused2         := 0;    // Unused
  //feb.checksum      := calc'd later (Same algorithm as RootBlock)
  //feb.data_blocks   := array[0..71] of longword - populated later

  for n := 0 to 45 do
    feb.info[n] := $00000000;       // array[0..45] of longword  (Unused)

  feb.unused3         := 0;         // Unused
  feb.parent := parentBlockNum;

  //feb.extension     :longword;    // Next file header extension block (0=this is the last)
  feb.sec_type        := $FFFFFFFD; // Secondary type : ST_FILE (= -3)

  FlipLong(feb.primtype);
  FlipLong(feb.header_key);
  FlipLong(feb.high_seq);
  Fliplong(feb.parent);
  Fliplong(feb.extension);
  Fliplong(feb.sec_type);

end;

{===============================================================================
 FUNCTION:    debugDispFEBInfo
 DESCRIPTION:

 ACCEPTS:     Debug function
              block number of parent block (File Header Block or File Extn Block)
              number of blocks to create
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.debugDispFEBInfo(cbf :Tcallbackfmt; blockNum :Longword);
var
  n     :Integer;
  feb   :^TFileExtnBlock;

begin

  feb := getFileHeaderBlockPtr(blockNum);

  cbf('');
  cbf('File Extn Block:');
  cbf('Primtype      = ' + formatLong('H', feb.primtype));
  cbf('Header Key    = ' + formatLong('H', feb.header_key) +' ('+formatLong('I', feb.header_key) +')');
  cbf('High Seq      = ' + formatLong('H', feb.high_seq)   +' ('+formatLong('I', feb.high_seq)   +')');
  cbf('Checksum      = ' + formatLong('H', feb.checksum));

  for n := 0 to 71 do
  begin
    cbf('Data block: '+IntToStr(n)+' = '+formatLong('H', feb.data_blocks[n])+'('+formatLong('I', feb.data_blocks[n]) +')');
  end;

  cbf('parent        = ' + formatLong('H', feb.parent) + ' ('+formatLong('I', feb.parent)+')');
  cbf('extension     = ' + formatLong('H', feb.extension) + ' ('+formatLong('I', feb.extension)+')');
  cbf('sec type      = ' + formatLong('H', feb.sec_type));
  cbf('');

end;


{===============================================================================
 FUNCTION:    createDataBlocks
 DESCRIPTION:

 ACCEPTS:
              block number of parent block (File Header Block or File Extn Block)
              number of blocks to create
 RETURNS:     Nothing
===============================================================================}
procedure TADFlib.createDataBlocks(
  cbf            :Tcallbackfmt;
  current_block  :Longword;
  numBlocks      :Integer;
  theData        :array of byte);

var
  n               :Integer;
  bytepos         :Integer;
  parentBlockNum  :Longword;
  ofsBlock        :^T_OFS_DataBlock;

begin

  cbf('Creating ' + IntToStr(numBlocks) + ' Data blocks');
  parentBlockNum := current_block;

  for n := 1 to numBlocks do
  begin

      // Start block number passed in should be the last block that was processed

      // Find next free block on disk
      current_block := block.getNextFreeBlock(current_block);

      // Processing for data block here....
      ofsBlock := getOFSDataBlockPtr(current_block);
      ofsBlock.primtype       := $00000008;
      Fliplong(ofsBlock.primtype);
      ofsBlock.header_key     := parentBlockNum;
      FlipLong(ofsBlock.header_key);
      ofsBlock.seq_num        := n;
      FlipLong(ofsBlock.seq_num);
      ofsBlock.data_size      := 488;
      FlipLong(ofsBlock.data_size);

      if (n = numBlocks) then
        ofsBlock.next_data := 0   // We are at last block
      else
        ofsBlock.next_data := block.getNextFreeBlock(current_block);

      FlipLong(ofsBlock.next_data);

      // Populate with data.
      // ofsBlock.data            :array[0..487] of byte; // 488 bytes data per block

      bytepos := 0;
      while (cur_byte < tot_bytes) do
      begin
        ofsBlock.data[bytepos] := theData[cur_byte];
        Inc(bytepos);
        Inc(cur_byte);

        if (bytepos >= 488) then
          break;
      end;

      // Calculate checksum for this data block once everything else filled in.
      ofsBlock.checksum     := block.CalcChecksum(diskMap[current_block*512],512,20,1);

      // DEBUG
      cbf('Data block ' + IntToStr(n) + ' stored at: ' + IntToStr(current_block));
      cbf('Primtype      = ' + formatLong('H', ofsBlock.primtype));
      cbf('Header Key    = ' + formatLong('H', ofsBlock.header_key)+' ('+formatLong('I', ofsBlock.header_key)+')');
      cbf('Seq Num       = ' + formatLong('H', ofsBlock.seq_num));
      cbf('Data Size     = ' + formatLong('H', ofsBlock.data_size)+' ('+formatLong('I', ofsBlock.data_size)+')');
      cbf('Next data     = ' + formatLong('H', ofsBlock.next_data)+' ('+formatLong('I', ofsBlock.next_data)+')');
      cbf('Checksum      = ' + formatLong('H', ofsBlock.checksum));
      cbf('');

      blockList.Add(IntToStr(current_block));
      block.markBlockAs(current_block, 'USED');

  end;
end;



{===============================================================================
 FUNCTION:    getByteAsChr
 DESCRIPTION: Returns byte at pos as a chr
 ACCEPTS:     Offset into diskMap array
 RETURNS:     Byte value as a chr
===============================================================================}
function TADFlib.getByteAsChr(pos: longword; source :String): String;
begin
  if source = 'diskMap' then
    Result := chr(diskMap[pos]);

  if source = 'fileMap' then
  begin
    if (pos < num_bytes_to_save) then
      Result := chr(fileMap[pos])
    else
      Result := ' ';
  end;
end;


{===============================================================================
 FUNCTION:    getByteAsHex
 DESCRIPTION: Returns byte at pos as two digit Hex value
 ACCEPTS:     Offset into diskMap array
 RETURNS:     Byte value as a Hex String
===============================================================================}
function TADFlib.getByteAsHex(pos: longword; source :String): String;
begin
  if source = 'diskMap' then
    Result := Format('%0.2x', [diskMap[pos]]);

  if source = 'fileMap' then
  begin
    if (pos < num_bytes_to_save) then
      Result := Format('%0.2x', [fileMap[pos]])
    else
      Result := ' ';

  end;
end;


procedure TADFlib.setByteInDiskMap(offset: Longword; byteval: byte);
begin
  diskMap[offset] := byteval;
end;

{===============================================================================
  FUNCTIONS IN THIS SECTION RETURN A POINTER TO A PARTICULAR STRUCTURE OF DATA

    function getDiskMapPtr()                          : pointer;
    function getBootBlockPtr()                        : pointer;
    function getRootBlockPtr()                        : pointer;
    function getFileHeaderBlockPtr(blockNum :Integer) : pointer;
    function getFileExtnBlockPtr(blockNum   :Integer) : pointer;
    function getOFSDataBlockPtr(blockNum    :Integer) : pointer;
    function getFFSDataBlockPtr(blockNum    :Integer) : pointer;
    function getBitmapBlockPtr(blockNum     :Integer) : pointer;
===============================================================================}

// Return diskmap ptr
function TADFlib.getDiskMapPtr(blockNum  :Integer): pointer;
var
  b   :pointer;
begin
  b := @diskMap[blockNum*512];
  result := b;
end;

// Return Bootblock ptr
function TADFlib.getBootBlockPtr(): pointer;
var
  b   :^TBootBlock;
begin
  b := @diskMap[0];
  Result := b;
end;

// Return Rootblock Ptr
function TADFlib.getRootBlockPtr(): pointer;
var
  b   :^TRootBlock;
begin
  b := @diskMap[880*512];
  Result := b;
end;

// Returns TFileHeaderblock pointer
function TADFlib.getFileHeaderBlockPtr(blockNum: Integer): pointer;
var
  b   :^TFileHeaderBlock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

// Returns TFileExtnBlock pointer
function TADFlib.getFileExtnBlockPtr(blockNum: Integer): pointer;
var
  b   :^TFileExtnBlock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

// Returns OFS datablock pointer
function TADFlib.getOFSDataBlockPtr(blockNum: Integer): pointer;
var
  b   :^T_OFS_DataBlock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

// Returns FFS datablock pointer
function TADFlib.getFFSDataBlockPtr(blockNum: Integer): pointer;
var
  b   :^T_FFS_DataBlock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

// Return bitmap block ptr
function TADFlib.getBitmapBlockPtr(blockNum: Integer): pointer;
var
  b   :^TBitmapblock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

// Return directory block ptr
function TADFlib.getDirectoryBlockPtr(blockNum: Integer): pointer;
var
  b   :^TUserDirectoryBlock;
begin
  b := @diskMap[blockNum*512];
  Result := b;
end;

function TADFlib.getByteFromDiskMap(pos: longword): byte;
var
  bytepos   :pointer;

begin
  bytepos := @diskMap[pos];
  result := byte(bytepos^);
end;

function TADFlib.getLongWordFromDiskMap(pos: longword): Longword;
var
  bytepos   :pointer;
begin
  bytepos := @diskMap[pos];
  Result := Longword(bytepos^);
end;

function TADFlib.getDiskMapPtr2(pos: Integer): pointer;
var
  b   :pointer;
begin
  b := @diskMap[pos];
  result := b;
end;


function TADFlib.isBootable: Boolean;
begin
  if ((block.getBlockChecksum(0) = block.calcBlockChecksum(0))
  and (getFileSystem = 'DOS')) then
    Result := true
  else
    Result := false;
end;

function TADFlib.getCRC32: Longword;
begin
  result := calcCRCFromMemory(getdiskMapPtr(0), 901120);
end;

function TADFlib.getFloppySize: Longint;
begin
  Result := fFloppySize;
end;

end.



