{  Unit name ...... SINFLATE (Secure Inflate)
   Version ........ 1.00
   Date ........... 2002-01-11
   Creator ........ Jgermeister Markus (jaegermeister.markus@gmx.de)

   Unit purpose ... This unit is intended to unpack data compressed with
                    the DEFLATE algorithm (used in PKZIP, GZIP and some
                    more) by easy to use procedure/function calls and
                    easy to understand sourcecode structures. Anyway, I'd
                    recommend to read the DEFLATE specifications which
                    can be found at the official GZIP website.
                        Without having read it first you probably won't
                    understand how this unit is working (except you know
                    a lot about huffman trees and LZ77 compression).
                        Although DEFLATE uses a combination of LZ77 with
                    huffman trees there are some special rules and of
                    course the structure how tree data are saved in DE-
                    FLATED data.
                        The difference between SINFLATE and INFLATE (by
                    Oliver Fromme) is that SINFLATE will take the main
                    program and the user care of errors, statistics and
                    of course how the unit works. This will result in
                    a little speed loss, but SINFLATE will NEVER cause
                    some runtime errors or will access memory outside
                    all allocated areas (useful for Virtual Pascal and
                    other than Turbo Pascal compilers). Apart from that
                    I tried to use understandable name definitions for
                    procedures/functions/variables and not two-character
                    names where nobody else like the programmer knows
                    what's their purpose.
                        Another thing is that the SI_Inflate procedure
                    will only read as much bits as needed, this is
                    a little bit slower (no lookup!) but the main
                    program won't need to care about if the decompressor
                    has read one or more bytes which don't belong to
                    the compressed data, so no more searching for
                    signatures or whatever for the main program.


   Conventions .... Variables which point to a certain type of data
                    get a lowercase "p" in front of the name,
                    types get a lowercase "t" in front.
                    Vars/Procs/Funcs used non-internally usable have
                    an "SI_" in front and are written uppercase.

   Initialization . Before decompressing some data you will need to do
                    two things:
                    1. Submit an SI_READ function which reads a byte
                           from the compressed data stream
                           (SI_READ:=MyRead;)
                    2. Submit an SI_WRITE procedure which takes a
                           datum X of type word as the first parameter
                           and writes or does something else with
                           the first X bytes of the array SI_WIND^,
                           which contains the uncompressed data.
                           Usually the word X will be $8000 until the
                           whole data has been decompressed.
                           (SI_WRITE:=MyWrite;)
                    An example can be seen some lines below! As you can
                    see in comparison with Oliver Fromme's inflate.pas
                    with this unit the user does not need to allocate
                    memory for the sliding window. As that amount of
                    memory always is needed for the inflate process
                    each time the SI_INFLATE is called it's first
                    reserved and after inflate is complete, it's
                    given free again.

   Copyrights ..... This unit and it's sourcecode aren't copyrighted.
                    You can do what you want with both, but you're welcome
                    if you mail me any bugs you've found, made improvements
                    or something else to the unit. Credits won't be missing! }

(*
  EXAMPLE MAIN PROGRAM USING THIS UNIT
  ------------------------------------

  uses SI_INFLATE;

  var input,output:file;
      inputbuffer:array[$0000..$1fff] of byte;  { 8kb read buffer for speed }
      inputbufferposition:word;
      inputbuffersize:word;

  procedure MyReadInitialize;
    begin
      inputbufferposition:=sizeof(inputbuffer);
    end;

  function MyRead:byte; far;        { far statement needed for Turbo Pascal }
    begin
      if inputbufferposition=sizeof(inputbuffer)    { if position is beyond }
        then begin                                      { the actual buffer }
          if filesize(input)-filepos(input)<sizeof(inputbuffer)  { read new }
            then inputbuffersize:=filesize(input)-filepos(input)    { block }
            else inputbuffersize:=sizeof(inputbuffer);        { and set the }
            inputbufferposition:=$0000;              { pointer to the start }
          blockread(input,inputbuffer,inputbuffersize);
        end;
      if inputbufferposition=inputbuffersize      { last byte has been read }
        then begin                           { but inflate still wants more }
          SI_BREAK:=TRUE;                  { then stop inflate by this line }
        end else begin
          MyRead:=inputbuffer[inputbufferposition];
          inc(inputbufferposition);
        end;
    end;

  procedure MyWrite(amount:word); far;
    begin                                { The ^ is needed for the case you }
      blockwrite(output,SI_WINDOW^,amount); { aren't familiar with pointers }
    end;

  begin
    MyReadInitialize;{ forces MyRead to read a new block for the first time }
    SI_READ:=MyRead;                  { assign your read function (NEEDED!) }
    SI_WRITE:=MyWrite;              { assign your write procedure (NEEDED!) }
    assign(input,'INPUT.DAT');               { standard file opening... :)) }
    assign(output,'OUTPUT.DAT');
    reset(input,1);
    rewrite(output,1);
    write('Inflating INPUT.DAT to OUTPUT.DAT... ');
    SI_INFLATE;                                               { Run inflate }
    writeln(SI_GETERROR);                   { SI_GETERROR returns a string! }
    close(input);
    close(output);
  end.
*)

UNIT SINFLATE;

INTERFACE

var SI_READ:function:byte;
    SI_WRITE:procedure(amount:word);

function SI_GETERROR:string;  { returns a text describing the occured error }
procedure SI_INFLATE;                          { starts the inflate process }

const SI_BREAK:boolean=false;
     { The break variable, set it to TRUE if you want to break the
       inflate process at any time (use it in SI_READ or SI_WRITE) }

type tWindow=array[$0000..$7fff] of byte;
    { The type for the uncompressed data array ("sliding" window), must
      be at least $8000 bytes long! }

var SI_WINDOW:^tWindow;
   { This is the type'd pointer variable for the sliding window, it must be
     allocated and deallocated by the main program }
    SI_POSITION:word;
   { The actual window position when increased from $7fff to $8000 then
     SI_WRITE($8000) is called and it's set back to $0000.
     Don't change it during the inflate process, just read it for
     further information on the process of inflate!!! }
    SI_ERROR:word;
   { This is the error indicator which is set when breaking an inflate
     process or when an error occured. The following value table let's you
     conclude what have been going wrong:
     SI_ERROR (word bitform hi-lo)  ee--|-----|--at|hhbb
         e = error occured          0   = no error
                                    1   = error during inflate
                                    2   = user break during inflate
         a = block header error     0   = no error
                                    1   = header corrupt (only types 0/2)
                                          on block type 0 this means
                                          that the checksum is wrong,
                                          on block type 2 this means
                                          one of the huffman tables
                                          it wrong, check t and h
         t = huffman error type     0   = incomplete
                                    1   = full
         h = huffman error tree     0   = no error
                                    1   = codelength tree (error)
                                    2   = literal/lengthcode tree (error)
                                    3   = distance tree (error)
         b = actual block type      0-2 = block type b
                                    3   = illegal block type (error) }


IMPLEMENTATION

function SI_GetError:string;
  var s:string;
  begin
    if SI_ERROR and $c000=$0000 then s:='okay' else
    if SI_ERROR and $4000=$4000 then s:='user break' else begin
      if SI_ERROR and $0020=$0020 then begin
        s:='block header corrupt';
        if SI_ERROR and $000c<>$0000 then begin
          s:=s+', ';
          if SI_ERROR and $000c=$0004 then s:=s+'lengthcode tree' else
          if SI_ERROR and $000c=$0008 then s:=s+'literal tree' else
          if SI_ERROR and $000c=$000c then s:=s+'distance tree';
          s:=s+' is ';
          if SI_ERROR and $0010=$0000 then s:=s+'incomplete' else s:=s+'full';
        end;
      end;
    end;
    SI_GetError:=s;
  end;

const SI_USERBREAK:word=$4000;

{  }
{  DEFLATE STATIC TABLES block  }
{  }

type tLengthcodeValueOffset      = array [$101..$11d] of word;
     tLengthcodeValueExtrabits   = array [$101..$11d] of byte;
     tDistancecodeValueOffset    = array [$000..$01d] of word;
     tDistancecodeValueExtrabits = array [$000..$01d] of byte;

var LengthcodeValueOffset      : ^tLengthcodeValueOffset;
    LengthcodeValueExtrabits   : ^tLengthcodeValueExtrabits;
    DistancecodeValueOffset    : ^tDistancecodeValueOffset;
    DistancecodeValueExtrabits : ^tDistancecodeValueExtrabits;

procedure AllocateStaticTables;
{ Allocates and generates the tables:
    - LengthcodeValueOffset
    - LengthcodeValueExtrabits
    - DistancecodeValueOffset
    - DistancecodeValueExtrabits }
  var Pos:byte;
      LengthcodeOffset,DistancecodeOffset:word;
      LengthcodeExtrabits,DistancecodeExtrabits:word;
  begin
    new(LengthcodeValueOffset);
    new(LengthcodeValueExtrabits);
    new(DistancecodeValueOffset);
    new(DistancecodeValueExtrabits);
    LengthcodeExtrabits:=0;
    DistancecodeExtrabits:=0;
    LengthcodeOffset:=$0003;
    DistancecodeOffset:=$0001;
    LengthcodeValueOffset^[$11d]:=$0102;
    LengthcodeValueExtrabits^[$11d]:=$0000;
    for pos:=$00 to $1d do begin
     { increase number of extra bits for length code table every 4th value }
      if (pos>=$08) and (pos and $03=$00) then inc(LengthcodeExtrabits);
     { increase number of extra bits for distance code table every 2nd value }
      if (pos>=$04) and (pos and $01=$00) then inc(DistancecodeExtrabits);
     { for pos<=$1c put value entry into length code table }
      if (pos<=$1b) then LengthcodeValueOffset^[pos+$101]:=LengthcodeOffset;
      if (pos<=$1b) then LengthcodeValueExtrabits^[pos+$101]:=LengthcodeExtrabits;
     { put value entry into distance code table }
      DistancecodeValueOffset^[pos+$000]:=DistancecodeOffset;
{      write(hexa(DistancecodeOffset,4),'/',hexa(DistancecodeExtrabits,1),'  ');}
      DistancecodeValueExtrabits^[pos+$000]:=DistancecodeExtrabits;
     { increase length and distance code values }
      inc(LengthcodeOffset,$0001 shl LengthcodeExtrabits);
      inc(DistancecodeOffset,$0001 shl DistancecodeExtrabits);
    end;
  end;

procedure DeAllocateStaticTables;
  begin
    dispose(LengthcodeValueOffset);
    dispose(LengthcodeValueExtrabits);
    dispose(DistancecodeValueOffset);
    dispose(DistancecodeValueExtrabits);
  end;


{  }
{  DEFLATE HUFFMAN TREE block   }
{  }

type tpHuffmanNode=^tHuffmanNode;
    { The pointer type for the following type }
     tHuffmanNode=record
    { The huffman-tree node structure, differs from common specifications }
                    value:word;
                   { The value of the acutal alphabet entry }
                    next:array[0..1] of tpHuffmanNode;
                   { The pointers to the children for both directions,
                     if NIL then there's no child for the actual direction }
                    endnode:array[0..1] of boolean;
                   { Tells us if the nodes in both directions either are
                     end nodes or not }
                  end;

procedure CreateNewHuffmanTree(var HuffmanTree:tpHuffmanNode);
{ Creates a virgin huffman tree root }
  var direction:byte;
  begin
     new(HuffmanTree);
     with HuffmanTree^ do begin
       value:=$ffff;
       for direction:=$00 to $01 do begin
         next[direction]:=NIL;
         endnode[direction]:=false;
       end;
     end;
  end;

procedure AddNewChildToHuffmanNode(var HuffmanNode:tpHuffmanNode; newdirection:byte; newvalue:word);
{ Adds a child in direction >newdirection< to the tree and gives the child
  the value >newvalue<. If >newvalue< is set between $0000 and $fffe the
  child is seen as an endnode, if $ffff the child is just another node with
  two children. }
  var newnode:tpHuffmanNode;
      direction:byte;
  begin
    new(newnode);
    with newnode^ do begin
      Value:=newvalue;
      for direction:=$00 to $01 do begin
        next[direction]:=NIL;
        endnode[direction]:=false;
      end;
    end;
    if newvalue<$ffff then HuffmanNode^.endnode[newdirection]:=TRUE;
    HuffmanNode^.next[newdirection]:=newnode;
  end;

function AddNewCodeToHuffmanTree(var HuffmanNode:tpHuffmanNode; Codelength:byte; Codevalue:word):boolean;
{ Tries to add a new codelength to the tree, but beware:
  It only returns FALSE if no space is found for a value! This means e.g.
  that if you have a complete huffman tree with all codelengths=8 and
  you try to add a value with codelength<8 it will return TRUE and writes
  the code as the bitstream 0000000 (always left direction).
  So you have to fulfill the following rules:
    - always add the values with short codelengths before you add
          the values with codelengths which are longer
    - to test if the tree is complete try to add another value
          with the highest codelength you've used yet for the tree,
          if TRUE is the result, the tree is incomplete,
          if FALSE is the result, the tree is complete }
  var resulttt:boolean;
  begin
    if Codelength=0 then exit;                      { impossible codelength }
    if not HuffmanNode^.endnode[0] then begin
      if HuffmanNode^.next[0]=NIL
        then AddNewChildToHuffmanNode(HuffmanNode,0,$ffff);
      if (Codelength=1) then begin
        HuffmanNode^.endnode[0]:=TRUE;
        HuffmanNode^.next[0]^.value:=Codevalue;
        resulttt:=TRUE;
      end else begin
        resulttt:=AddNewCodeToHuffmanTree(HuffmanNode^.next[0],Codelength-1,Codevalue);
      end;
{      if result then write(0);}
    end else resulttt:=FALSE;
    if resulttt=false then if not HuffmanNode^.endnode[1] then begin
      HuffmanNode^.endnode[0]:=TRUE;
      if HuffmanNode^.next[1]=NIL
        then AddNewChildToHuffmanNode(HuffmanNode,1,$ffff);
      if (Codelength=1) then begin
        HuffmanNode^.endnode[1]:=TRUE;
        HuffmanNode^.next[1]^.value:=Codevalue;
        resulttt:=TRUE;
      end else begin
        resulttt:=AddNewCodeToHuffmanTree(HuffmanNode^.next[1],Codelength-1,Codevalue);
      end;
{      if result then write(1);}
    end else begin
      resulttt:=FALSE;
      HuffmanNode^.endnode[1]:=TRUE;
    end;
    AddNewCodeToHuffmanTree:=resulttt;
  end;

procedure FreeHuffmanTree(var HuffmanNode:tpHuffmanNode);
  begin
    if HuffmanNode^.next[0]<>NIL then FreeHuffmanTree(HuffmanNode^.next[0]);
    if HuffmanNode^.next[1]<>NIL then FreeHuffmanTree(HuffmanNode^.next[1]);
    dispose(HuffmanNode);
  end;

{  }
{  BITWISE READ block  }
{  }

var BitBuffer:byte;
    BitNumber:byte;

function ReadBits(NumberOfBits:byte):word;
{ Reads >NumberOfBits< bits from the input stream and returns it/them
  as a longint. This causes a little speed decrease but assures that
  not too much bytes are read and the caller doesn't have to calculate
  the byte/word/integer/whatever value by hand. }
var ResultMask:word;
    Resulttt:word;
begin
  ResultMask:=1;
  Resulttt:=0;
  while NumberOfBits>0 do begin
    if BitNumber=8 then begin
      BitNumber:=0;
      BitBuffer:=SI_READ;
    end;
    inc(Resulttt,(BitBuffer and $01)*ResultMask);
    ResultMask:=ResultMask shl $01;
    BitBuffer:=BitBuffer shr $01;
    inc(BitNumber);
    dec(NumberOfBits);
  end;
  ReadBits:=Resulttt;
end;

function ReadBit:byte;
{ Reads one bits from the input stream and returns it/them
  as a byte. }
begin
  if BitNumber=8 then begin
    BitNumber:=0;
    BitBuffer:=SI_READ;
  end;
  ReadBit:=BitBuffer and $01;
  BitBuffer:=BitBuffer shr $01;
  inc(BitNumber);
end;

{  }
{  HUFFMAN-TREE-USING block  }
{  }

const CodelengthOrder:array [$00..$12] of byte=
            ($10,$11,$12,$00,$08,$07,$09,$06,$0a,$05,$0b,$04,$0c,$03,$0d,$02,
             $0e,$01,$0f);

type tAlphabetCodelength         = array [$000..$13d] of byte;
    { where $000-$0ff is for literals, $100 for the end of block sign,
      $101-$11d for the length codes and $11e-$13b for the distance codes }
     tCodelengthCodelength       = array [$000..$012] of byte;

var AlphabetCodelength         : ^tAlphabetCodelength;
    LiteralHuffmanTree         : tpHuffmanNode;
    DistanceHuffmanTree        : tpHuffmanNode;
    CodelengthCodelength       : ^tCodelengthCodelength;
    CodelengthHuffmanTree      : tpHuffmanNode;

procedure CreateStaticCodeTrees;
var Value:word;
  begin
    CreateNewHuffmanTree(LiteralHuffmanTree);
    for Value:=$100 to $117 do AddNewCodeToHuffmanTree(LiteralHuffmanTree,7,Value);
    for Value:=$000 to $08f do AddNewCodeToHuffmanTree(LiteralHuffmanTree,8,Value);
    for Value:=$118 to $11f do AddNewCodeToHuffmanTree(LiteralHuffmanTree,8,Value);
    for Value:=$090 to $0ff do AddNewCodeToHuffmanTree(LiteralHuffmanTree,9,Value);
    CreateNewHuffmanTree(DistanceHuffmanTree);
    for Value:=$00 to $1f do AddNewCodeToHuffmanTree(DistanceHuffmanTree,5,Value);
  end;

procedure FreeStaticCodeTrees;
  begin
    FreeHuffmanTree(LiteralHuffmanTree);
    FreeHuffmanTree(DistanceHuffmanTree);
  end;

function DecodeValue(ActualNode:tpHuffmanNode):word;
  begin

    repeat
      ActualNode:=ActualNode^.next[ReadBit];
    until (not (ActualNode^.Value=$ffff)) or (SI_BREAK);

    if SI_BREAK then SI_ERROR:=$4000;
    DecodeValue:=ActualNode^.Value;
  end;

procedure ReadDynamicCodeTrees(CodelengthNumber,LiteralNumber,DistanceNumber:word);
  var CodeValue:word;
      Lengthcode,CodeLength:byte;
      RepeatAmount:word;
      RepeatValue:byte;
      ActualCodeLength,HighestCodeLength,Actual2,Highest2:byte;
      BuildSuccess:boolean;
  begin
   { Read codelength codelengths (first tree) }
    new(CodelengthCodelength);
    fillchar(CodelengthCodelength^,$12,$00);
    HighestCodeLength:=$00;
    for CodeValue:=$00 to CodelengthNumber-$01 do if not SI_BREAK then begin
      ActualCodeLength:=ReadBits(3);
      if ActualCodeLength>HighestCodeLength then HighestCodeLength:=ActualCodeLength;
      CodelengthCodelength^[CodelengthOrder[CodeValue]]:=ActualCodeLength;
    end;
    if SI_BREAK then begin SI_ERROR:=SI_USERBREAK; dispose(CodelengthCodelength); Exit; end;
   { Build up tree }
    CreateNewHuffmanTree(CodelengthHuffmanTree);
    for CodeLength:=$01 to $0f do begin
      for CodeValue:=$00 to $12 do if CodeLength=CodelengthCodelength^[CodeValue]
        then BuildSuccess:=AddNewCodeToHuffmanTree(CodelengthHuffmanTree,CodeLength,CodeValue);
    end;
    dispose(CodelengthCodelength);
    if BuildSuccess then begin
      BuildSuccess:=not AddNewCodeToHuffmanTree(CodelengthHuffmanTree,HighestCodeLength,0);
      if not BuildSuccess then begin
        SI_ERROR:=(SI_ERROR and $7fc3)+$8024;
        Exit;
      end;
    end else begin
      SI_ERROR:=(SI_ERROR and $7fc3)+$8034;
      Exit;
    end;
   { Real literal + distance (alphabet) codelengths }
    new(AlphabetCodelength);
    RepeatAmount:=0;
    HighestCodelength:=0;
    Highest2:=0;
    for CodeValue:=$000 to LiteralNumber+DistanceNumber-$01 do begin
      if not SI_BREAK then begin
        if RepeatAmount=0 then begin
          Lengthcode:=DecodeValue(CodelengthHuffmanTree);
          if Lengthcode<$10 then begin
            AlphabetCodelength^[CodeValue]:=Lengthcode;
            if CodeValue<LiteralNumber then begin
              if Lengthcode>HighestCodeLength then HighestCodeLength:=LengthCode;
            end else begin
              if Lengthcode>Highest2 then Highest2:=Lengthcode;
            end;
          end else
          if Lengthcode=$10 then begin
            RepeatAmount:=$02+ReadBits(2);
            RepeatValue:=AlphabetCodelength^[CodeValue-$01];
            AlphabetCodelength^[CodeValue]:=RepeatValue;
          end else
          if lengthcode=$11 then begin
            RepeatAmount:=$02+ReadBits(3);
            RepeatValue:=$00;
            AlphabetCodelength^[CodeValue]:=RepeatValue;
          end else
          if lengthcode=$12 then begin
            RepeatAmount:=$0a+ReadBits(7);
            RepeatValue:=$00;
            AlphabetCodelength^[CodeValue]:=RepeatValue;
          end;
          if SI_BREAK then begin
            SI_ERROR:=SI_USERBREAK;
            dispose(AlphabetCodelength);
            FreeHuffmanTree(CodelengthHuffmanTree);
            Exit;
          end;
        end else begin
          AlphabetCodelength^[CodeValue]:=RepeatValue;
          dec(RepeatAmount);
        end;
      end;
    end;
   { Free huffman tree for codelength resolving (not needed anymore) }
    FreeHuffmanTree(CodelengthHuffmanTree);
   { Build up literal tree }
    CreateNewHuffmanTree(LiteralHuffmanTree);
    for CodeLength:=$01 to $0f do begin
      for CodeValue:=$000 to LiteralNumber-1 do if CodeLength=AlphabetCodelength^[CodeValue] then
        BuildSuccess:=AddNewCodeToHuffmanTree(LiteralHuffmanTree,CodeLength,CodeValue);
    end;
    if BuildSuccess then begin
      BuildSuccess:=not AddNewCodeToHuffmanTree(LiteralHuffmanTree,HighestCodeLength,0);
      if not BuildSuccess then begin
        SI_ERROR:=(SI_ERROR and $7fc3)+$8028;
        FreeHuffmanTree(LiteralHuffmanTree);
        dispose(AlphabetCodelength);
        Exit;
      end;
    end else begin
      SI_ERROR:=(SI_ERROR and $7fc3)+$8038;
      FreeHuffmanTree(LiteralHuffmanTree);
      dispose(AlphabetCodelength);
      Exit;
    end;
   { Build up distance tree }
    CreateNewHuffmanTree(DistanceHuffmanTree);
    for CodeLength:=$01 to $0f do begin
      for CodeValue:=$00 to DistanceNumber-1 do if CodeLength=AlphabetCodelength^[CodeValue+LiteralNumber]
        then BuildSuccess:=AddNewCodeToHuffmanTree(DistanceHuffmanTree,CodeLength,CodeValue);
    end;
    if BuildSuccess then begin
      BuildSuccess:=not AddNewCodeToHuffmanTree(DistanceHuffmanTree,Highest2,0);
      if not BuildSuccess then begin
        SI_ERROR:=(SI_ERROR and $7fc3)+$802c;
        FreeHuffmanTree(DistanceHuffmanTree);
        dispose(AlphabetCodelength);
        Exit;
      end;
    end else begin
      SI_ERROR:=(SI_ERROR and $7fc3)+$803c;
      FreeHuffmanTree(DistanceHuffmanTree);
      dispose(AlphabetCodelength);
      Exit;
    end;
    dispose(AlphabetCodeLength);
  end;

procedure FreeDynamicCodeTrees;
  begin
    FreeHuffmanTree(LiteralHuffmanTree);
    FreeHuffmanTree(DistanceHuffmanTree);
  end;

{  }
{  DEFLATE-BLOCK-HANDLING block  }
{  }

const LiteralAlphabetLengthOffset    : word = $0101;
      DistanceAlphabetLengthOffset   : byte =   $01;
      LengthcodeAlphabetLengthOffset : byte =   $04;

procedure OutputByte(b:byte);
begin
  SI_WINDOW^[SI_POSITION]:=b;
  inc(SI_POSITION);
  if SI_POSITION=$8000 then begin
    SI_WRITE($8000);
    SI_POSITION:=$0000;
  end;
end;

procedure CopyBytes(distance,length:word);
begin
  while length>0 do begin
    OutputByte(SI_WINDOW^[(SI_POSITION+$8000-Distance) and $7fff]);
{    OutputByte($ff);}
    dec(length);
  end;
end;

procedure SI_INFLATE;
var LastBlock,BlockType:byte;
    CodelengthNumber,LiteralNumber,DistanceNumber:word;
    Literal:word;
    Length:word;
    Distance:word;
    Stored:pointer;
  begin
   { Force ReadBits to read a byte first }
    BitNumber:=8;
   { Set sliding window position to the start }
    SI_POSITION:=0;
    SI_ERROR:=$0000;
    AllocateStaticTables;
    new(SI_WINDOW);
    repeat
      LastBlock:=ReadBit;
      BlockType:=ReadBits(2);
{         writeln('LastBlock              = $',hexa(LastBlock,2));
           write('BlockType              = $',hexa(BlockType,2),' (');
      if BlockType=0 then write('stored') else
      if BlockType=1 then write('static huffman') else
      if BlockType=2 then write('dynamic huffman') else begin
        write('illegal');
        SI_ERROR:=$8003;
      end;
      writeln(')');}

     { Decompress the specific block }
      if BlockType=0 then begin
       { ignore all bits till next byte }
        BitNumber:=8;
        Length:=SI_READ*$100+SI_READ;
        Distance:=SI_READ*$100+SI_READ;
{        writeln('Length                 = $',hexa(Length,4));
        writeln('Length complement      = $',hexa(Distance,4));}
        if (Length xor Distance)<>$ffff then SI_ERROR:=$8020 else
        while (Length>0) and (SI_ERROR and $c000=$0000) do begin
          Literal:=SI_READ;
          OutputByte(Literal);
          dec(Length);
        end;
      end else
      if BlockType=1 then begin
        CreateStaticCodeTrees;
        if SI_ERROR and $8000=$0000 then
        repeat
          Literal:=DecodeValue(LiteralHuffmanTree);
          if (SI_ERROR and $c000<>$0000) then Literal:=$100 else begin
            if Literal<$100 then OutPutByte(Literal) else
            if Literal=$100 then else
            if Literal<=$11d then begin
              Length:=LengthcodeValueOffset^[Literal]+ReadBits(LengthcodeValueExtrabits^[Literal]);
              Distance:=DecodeValue(DistanceHuffmanTree);
              if Distance>$1d then begin
                SI_ERROR:=$8000;
                Literal:=$100;
              end else begin
                Distance:=DistancecodeValueOffset^[Distance]+ReadBits(DistancecodeValueExtrabits^[Distance]);
                CopyBytes(Distance,Length);
              end;
            end else begin SI_ERROR:=$8000; Literal:=$100; end;
          end;
        until (Literal=$100);
        FreeStaticCodeTrees;
      end else
      if BlockType=2 then begin
        LiteralNumber:=LiteralAlphabetLengthOffset+ReadBits(5);
        DistanceNumber:=DistanceAlphabetLengthOffset+ReadBits(5);
        CodelengthNumber:=LengthCodeAlphabetLengthOffset+ReadBits(4);
        ReadDynamicCodeTrees(CodelengthNumber,LiteralNumber,DistanceNumber);
{        writeln('LiteralNumber          = $',hexa(LiteralNumber,4));
        writeln('DistanceNumber         = $',hexa(DistanceNumber,4));
        writeln('CodelengthNumber       = $',hexa(CodelengthNumber,4));
        writeln('DynamicCodeTreesCheck  = $',hexa(SI_ERROR,4)+' ('+SI_GetError+')');}
        if SI_ERROR and $8000=$0000 then
        repeat
          Literal:=DecodeValue(LiteralHuffmanTree);
          if (SI_ERROR and $c000<>$0000) then Literal:=$100 else begin
            if Literal<$100 then OutPutByte(Literal) else
            if Literal=$100 then else
            if Literal<=$11d then begin
              Length:=LengthcodeValueOffset^[Literal]+ReadBits(LengthcodeValueExtrabits^[Literal]);
              Distance:=DecodeValue(DistanceHuffmanTree);
              Distance:=DistancecodeValueOffset^[Distance]+ReadBits(DistancecodeValueExtrabits^[Distance]);
{              if SI_POSITION div $100=$12 then begin
                write(hexa(Distance,4),'/',hexa(Length,4),'/',hexa(SI_POSITION,4),'  ');
              end;}
              CopyBytes(Distance,Length);
            end else begin SI_ERROR:=$8000; Literal:=$100; end;
          end;
        until (Literal=$100);
        if (SI_ERROR and $8000=$0000) and (SI_ERROR<>$8000) then FreeDynamicCodeTrees;
      end;
    until LastBlock=1;
    if (SI_ERROR and $c000=$0000) and (SI_POSITION>$0000) then SI_WRITE(SI_POSITION);
    DeAllocateStaticTables;
    dispose(SI_WINDOW);
  end;

begin end.
